#include <stdio.h>
#include <bench.h>

/* CRC_CALC -- calculate a standard crc checksum
 *
 * Calculate the 16-bit CRC of a string using a byte-oriented tableless
 * algorithm invented by Andy Lowry (Columbia University).  The magic number
 * 010201 is derived from the CRC-CCITT polynomial x^16+x^12+x^5+1.
 *
 * Note - this function has been adapted for a file checksum by doing
 *        the fget's in the calculation.
 *        Also, parity is always 0 (no parity) right now.  I'll get to
 *        this later (if necessary).
*/
long crc_calc(Fp)
FILE *Fp;
{
	long crc;
	register unsigned int c, q;
	int parity = 0;

	crc = 0L;

	while((c = getc(Fp)) != EOF)					/* get each character */
	{
		if (c == '\n')
			continue;
		if (parity)
			c &= 0177;
		else
			c &= 0377;
		q = (unsigned int) (crc ^ c) & 0x0F;			/* Low-order nibble */
		crc = (crc >> 4) ^ (q * 0x1081);
		q = (unsigned int) (crc ^ (c >> 4)) & 0x0F;	/* High order nibble */
		crc = (crc >> 4) ^ (q * 0x1081);
	}
	return(crc);
}

