/* ==( io/src/ibtrieve.c )== */

/* ----------------------------------------------- */
/* Pro-C  Copyright (C) 1988 - 1990 Vestronix Inc. */
/* Modification to this source is not supported    */
/* by Vestronix Inc.                               */
/*            All Rights Reserved                  */
/* ----------------------------------------------- */
/* Written   AMcDonald & BZigon                    */
/* Modified  VV  28-Mar-90  See comments below     */
/* ----------------------------------------------- */
/* %W%  (%H% %T%) */

/*
 *  Modifications
 *
 *  28-Mar-90  VvA - adjustments for locking modes, transactions
 *  14-Mar-90  VvA - Locking and Indexing adjustment
 *  01-Feb-90  VvA - Enabled openmode in open_file()
 *  14-Dec-89  VvA - V.2 modifications
 *  01-Aug-89  VvA - Modified for generic I/O interface
 *  01-Sep-88  VvA - Numerical and segmented keys added
 *  01-Jul-88  Nig - Adapted for V4.11
*/

/*  Btrieve specific I/O calls via general IOGEN.C interface  */


/*****************************************************************************
 *                                                                           *
 *  This program handles the interface between i/o calls from the generic    *
 *  I/O module and Btrieve.  Note that Btrieve does not use file descriptors *
 *  therefore we have to take care of them here.                             *
 *                                                                           *
 *  Btrieve Calling format :                                                 *
 *                                                                           *
 *  BTRV(op, pos_blk, data_buf, &rec_len, key_buf, key_num)                  * 
 *                                                                           *
 *     int op;              operation code                                   *
 *     char *pos_blk;       position block                                   *
 *     char *data_buf;      data buffer                                      * 
 *     int  &rec_len;       record length                                    *
 *     char *key_buf;       key buffer                                       *
 *     int  key_num;        key number                                       *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

# include <stdio.h>
# include <iodef.h>
# include <iomsg.h>
# include <proc.io>
# include <bench.h>
# include <iosup.h>


/*
* Btrieve Definitions
*/
#define       B_OPEN           0
#define       B_CLOSE          1
#define       B_INSERT         2
#define       B_UPDATE         3
#define       B_DELETE         4
#define       B_GETEQUAL       5
#define       B_GETNEXT        6
#define       B_GETPREV        7
#define       B_GETGRTorEQUAL  9
#define       B_GETLOWEST     12
#define       B_GETLAST       13
#define       B_CREATE        14
#define       B_STAT          15
#define       B_BEG_TRN       19
#define       B_END_TRN       20
#define       B_ABT_TRN       21
#define       B_GETPOSN       22
#define       B_GETDIRECT     23
#define       B_ULOCK         27
#define       B_LOCK         100
#define		  B_NWLOCK		  200


/*
 * Values to define key attributes
*/
#define       DUP              1	/* allows duplicates */
#define       MOD              2	/* able to change key */
#define       BIN              4	/* binary */
#define       NL               8	/* null */
#define       SEG             16	/* segmented */
#define       ALT             32	/* alternative collating sequence */
#define       DES             64        /* Descending */
#define       SUP            128        /* Supplemental */
#define       EXT            256        /* Extended Type */


/*
 * Different Key Types
*/
#define       B_STRING           0
#define       B_INTEGER          1
#define       B_FLOAT            2
#define       B_DATE             3
#define       B_TIME             4
#define       B_DECIMAL          5
#define       B_MONEY            6
#define       B_LOGICAL          7
#define       B_NUMERIC          8
#define       B_BFLOAT           9
#define       B_LSTRING         10
#define       B_ZSTRING         11
#define       B_UBINARY         14


/*
 * Modes for Btrieve file open calls
*/
#define       FAST            -1	/* increases performance by disabling */
                                    /* automatic data recovery capability */
#define       RECOV           -2	/* to open a damaged file which can't */
                                    /* be recovered automatically	      */
#define       EXCL            -4    /* exclusive use on shared device     */
#define       NORMAL           0


/*
 * Miscellaneous defines
*/
#define       FCB_SIZE       128	/* needed for file control buffer  */


/*
 * Btrieve Error Codes
*/
#define       BTBADOPEN        3
#define       BTKEYNFOUND      4
#define       BTDUP            5         /* where no duplicates are allowed */
#define       BTINDEXPOSN      7
#define       BTNO_MORE_RECS   9
#define       BTNO_FILE       12              /* probably need to create it */
#define       BTNO_BTRIEVE    20                     /* btrieve not started */
#define       BTTRANSERR      36
#define       BTTRANSACTV     37
#define       BTTRANSFIO      38
#define       BTTRANSEND      39
#define       BTTRANSMAX      40
#define		  BTCREATE_ERR    25
#define       BTTOOMANYKEY    73
#define       BTRECLEN        74
#define       BTRECINUSE      84
#define		  BFILEINUSE      85          /* file already opened and locked */


#ifdef UNIX
# define BTRV btrv
#endif


/* Function prototypes */
# ifdef ANSI
static int i_addrec(int, char *);
static int i_close_file(int, char *);
static int i_commit(int, char *);
static int i_delrec(int, char *);
static int i_filename(int, char *);
static int i_findkey(int, char *);
static int i_firstkey(int, char *);
static int i_init_file(int, char *);
static int i_lastkey(int, char *);
static int i_lockrec(int, char *);
static int i_login(int, char *);
static int i_logoff(int, char *);
static int i_nextrec(int, char *);
static int i_open_file(int, char *);
static int i_prevrec(int, char *);
static int i_rereadrec(int, char *);
static int i_rollback(int, char *);
static int i_selectinx(int, char *);
static int i_transact(int, char *);
static int i_unlock_rec(int, char *);
static int i_updrec(int, char *);
static int create_file(int, int);
static int io_xlate(int, int, char *);
int  BTRV(int, char *, char *, int *, char *, int);
# else
static int i_addrec();
static int i_close_file();
static int i_commit();
static int i_delrec();
static int i_filename();
static int i_findkey();
static int i_firstkey();
static int i_init_file();
static int i_lastkey();
static int i_lockrec();
static int i_login();
static int i_logoff();
static int i_nextrec();
static int i_open_file();
static int i_prevrec();
static int i_rereadrec();
static int i_rollback();
static int i_selectinx();
static int i_transact();
static int i_unlock_rec();
static int i_updrec();
static int create_file();
static int io_xlate();
int  BTRV();
# endif


/*
 *  Btrieve structure tag definitions
*/
struct key_def
{
   int kpos;          /* offset of key into rec (starting at 1) */
   int klen;                      /* key length                 */
   int kflag;                     /* flag modes eg duplicate ?  */
   char not_used_k[4];
   char ktype;
   char reserved_k[5];
};

struct file_def
{
   int reclen;
   int page_size;
   int no_keys;
   char not_used_f[4];
   int file_flag;
   char reserved_k[2];
   int pre_alloc;
   struct key_def key[MAX_KEYS];
};


/*
 * structure for Btrieve-specific values 
*/
struct btr_def
{
   char *fcb;                                /* Btrieve file control buffer */
   char *key_buf;                                     /* Btrieve key buffer */
   long recposn;
};


static int BT_LOCK = 0;          /* = B_LOCK if wait, = B_NWLOCK if no wait */
static int BT_TRANS = FALSE;              /* if transaction processing used */
static int posnlen = 4;
static struct btr_def fbtr[MAX_FILES];



/*
 *
 * Interface Functions
 *
*/


/*
 *  Sets up File Name extension.
*/
static int i_filename(fd_sys, buffer)
int fd_sys;
char *buffer;
{
#ifdef MSDOS 
   char *tok;
#endif

   if(strchr(buffer, '.') == NULL)   /* use extension .DAT if none supplied */
      strcat(buffer, ".DAT");

#ifdef MSDOS 
   if ((tok = strchr(buffer, 47)) != NULL)       /* frontslash to backslash */
      *tok = (char)92;
#endif

   return(IOGOOD);
}

/*
 *  File initialization
*/
static int i_init_file(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  File open function
*/
static int i_open_file(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int n, m, stat, filsiz, tlen, max_key_len;
   int ko, key_overhead = 0;
   int BT_OPEN;
   struct file_def fil;                                     /* from Btrieve */
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];

   if (!(bptr->fcb = alloc(FCB_SIZE)))       /* Btrieve file control buffer */
   {
      errmsg(FileNoMemory);                     /* not enough mem available */
      fptr->active = NO;
      return(IOERROR);
   }

   n = 0;                              /* calculations for Btrieve pagesize */
   while (fptr->keys[n].segcount != -1)
   {
      tlen = 0;
      for (m = 0; m < fptr->keys[n].segcount; m++)    /* length of all segs */
         tlen += fptr->keys[n].seglen[m];

      if (tlen > max_key_len)                /* maximum key length required */
         max_key_len = tlen;  

      if ((ko = (tlen + (fptr->keys[n].keytype == KEY_DUPLICATE ? 8 : 0 ))) > key_overhead)
         key_overhead = ko;        /* max keylen incl 8 bytes for duplicate */
      n++;
   }
   key_overhead += 6;                     /* add fixed Btrieve key overhead */

   if (!(bptr->key_buf = alloc(max_key_len)))
   {
      errmsg(FileNoMemory);                     /* not enough mem available */
      free(bptr->fcb);
      fptr->active = NO;
      return(IOERROR);
   }

   if (fptr->openmode & OUTPUT_FLAG)
      unlink(fptr->filname);            /* for OUTPUT delete existing files */

   BT_OPEN = FAST;                  /* default open mode, exceptions follow */
   if (fptr->openmode & INPUT_FLAG)
      BT_OPEN = RECOV;
   if (fptr->openmode & P_EXCLUSIVE)
      BT_OPEN = EXCL;

   if (fptr->openmode & P_TRANSACT)
      BT_TRANS = TRUE;

   if (stat = BTRV(B_OPEN, (char *)bptr->fcb, NULL, &fptr->rec_len, fptr->filname, BT_OPEN))
   {
      if (stat == BTNO_BTRIEVE)
      {
         fptr->active = NO;
         free(bptr->key_buf);
         free(bptr->fcb);
         errmsg(FileEngNotStart_s, "Btrieve");
         return(IOERROR);
      }

      if (stat == BTNO_FILE && !(fptr->openmode & INPUT_FLAG))
      {                                           /* try to create the file */
         if (create_file(fd_sys, key_overhead) == IOERROR)
            return(IOERROR);

         if (stat = BTRV(B_OPEN, (char *)bptr->fcb, NULL, &fptr->rec_len, fptr->filname, BT_OPEN))
            return(io_xlate(fd_sys, stat, "BTRV OPEN1"));
      }
      else
         return io_xlate(fd_sys, stat, "BTRV OPEN2");
   }
   else                   /* at this point, the file is successfully opened */
   {
      filsiz = sizeof(fil);

      if (stat = BTRV(B_STAT, (char *)bptr->fcb, (char *)&fil, &filsiz, bptr->key_buf, 0))
         errmsg(FileGenError_sds, "Btrieve", stat, "LENCHK");

      if (fil.reclen != fptr->rec_len)
      {
         i_close_file(fd_sys, buffer);
         errmsg(FileRecLenChg_s, fptr->filname);
         return(IOERROR);
      }
   }

   return(IOGOOD);
}


/*
 * Btrieve create file function - includes page size calculations
 * - when this routine is finished, the file still needs to be opened
*/
static int create_file(fd_sys, maxkey_len)
int fd_sys;
int maxkey_len;      /* this includes fixed 6 and duplicate 8 byte overhead */
{
   int j, n, m, fx, stat;
   int filsiz, treclen, dups = 0, pagesize = 512;
   struct file_def fil;
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];

   fil.reclen = fptr->rec_len;

   j = n = 0;                                          /* load the key info */
   while (fptr->keys[n].segcount != -1)
   {                                     /* max key length and key overhead */
      for (m = 0; m < fptr->keys[n].segcount; m++)
      {
         fil.key[j].kpos = fptr->keys[n].segstart[m] + 1;
         fil.key[j].klen = fptr->keys[n].seglen[m];

         if (fptr->keys[n].keytype == KEY_DUPLICATE)
            fil.key[j].kflag = DUP | MOD | EXT;
         else
            fil.key[j].kflag = MOD | EXT;

         fx = fptr->keys[n].fldindex[m];
         switch(fptr->flds[fx].fldtype)         /* data type of key segment */
         {
            case(CHRTYP) :
               fil.key[j].ktype = B_STRING;
               break;
            case(INTTYP) :
            case(LNGTYP) :
               fil.key[j].ktype = B_INTEGER;
               break;
            case(FLTTYP) :
            case(DBLTYP) :
               fil.key[j].ktype = B_FLOAT;
            case(DATTYP) :
               if (fptr->keys[n].seglen[m] == 6)          /* date as YYMMDD */
                  fil.key[j].ktype = B_STRING;
               else                                     /* date as long int */
                  fil.key[j].ktype = B_INTEGER;
         }

         if (fptr->keys[n].segcount > m+1)       /* another segment follows */
            fil.key[j].kflag = fil.key[j].kflag | SEG;

         j++;                          /* increment total key segment count */
      }
      
      if (fptr->keys[n].keytype == KEY_DUPLICATE)
         dups++;                     /* accumulate number of duplicate keys */
         
      n++;                                     /* increment total key count */
   }
   
   if (j > MAX_KEYS)                      /* exceeded Btrieve key+seg limit */
   {
      fptr->active = NO;
      free(bptr->key_buf);
      free(bptr->fcb);
      return(io_xlate(fd_sys, BTTOOMANYKEY, "BTRV CRT1"));
   }

   fil.no_keys = n;                      /* total number of keys (not segs) */
   fil.file_flag = 0;

/*
 * Page size is calculated as follows:
 *     PAGESIZE  =  Record Length + 6 + num of Dupl.Keys * 8
 * This value must be a multiple of 512 and less than or equal to 4096, 
 * and must be larger than  8 * (Longest Key + Key Overhead)
 */

   treclen = fptr->rec_len + 6 + (dups * 8);

   while (pagesize <= 4096)
   {
      while (pagesize <= 4096 && pagesize < treclen)
         pagesize += 512;
      if (pagesize < (8 * maxkey_len))
         pagesize += 512;
      else
         break;
   }

   if (pagesize > 4096)                   /* Btrieve page size limit (4096) */
   {
      fptr->active = NO;
      free(bptr->key_buf);
      free(bptr->fcb);
      return(io_xlate(fd_sys, BTRECLEN, "BTRV CRT2"));
   }

   if (pagesize == 512 && j > 8)                    /* if page size is 512, */
      pagesize = 1024;                    /* number of keys is limited to 8 */

   fil.page_size = pagesize;
   filsiz = sizeof(fil);

   if (stat = BTRV(B_CREATE, (char *)bptr->fcb, (char *)&fil, &filsiz, fptr->filname, 0))
      return(io_xlate(fd_sys, stat, "BTRV CRT3"));
   else
      return(IOGOOD);
}


/*
 *  File close function
*/
static int i_close_file(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];

   if (stat = BTRV(B_CLOSE, (char *)bptr->fcb, NULL, &fptr->rec_len, NULL, 0))
      return(io_xlate(fd_sys, stat, "BTRV CLOSE"));
	
   free(bptr->fcb);
   free(bptr->key_buf);

   return(IOGOOD);
}

/*
 *  Select an index to perform processing on - already done in IOGEN.C
*/
static int i_selectinx(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   if (fd[fd_sys].lockmode == NOLOCK)
      BT_LOCK = NOLOCK;
/*                --- since wait lock implemented in iogen.c, don't do it here
   else if (fd[fd_sys].lockmode == INFLOCK)
      BT_LOCK = B_LOCK;
*/
   else
      BT_LOCK = B_NWLOCK;
   fd[fd_sys].got_rec = NO;
   return(IOGOOD);
}


/*
 *  Find a record by key value
*/
static int i_findkey(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int m, ko, BT_OP, stat;
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];
   struct keyinfo *kptr = &fd[fd_sys].keys[fd[fd_sys].cur_key];

   for (m = 0, ko = 0; m < kptr->segcount; m++)
   {                               /* copy each segment into the key buffer */
      bytecpy(&bptr->key_buf[ko], buffer+kptr->segstart[m], kptr->seglen[m]);
      ko += kptr->seglen[m];
   }

   BT_OP = BT_LOCK + (fptr->exact ? B_GETEQUAL : B_GETGRTorEQUAL);

   if (stat = BTRV(BT_OP, (char *)bptr->fcb, buffer, &fptr->rec_len, bptr->key_buf, fptr->cur_key))
      return((!fptr->exact && (stat == BTNO_MORE_RECS)) ? IOEOF : io_xlate(fd_sys, stat, "BTRV FND"));

   if (stat = BTRV(B_GETPOSN, (char *)bptr->fcb, (char *)&bptr->recposn, &posnlen, NULL, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV FINDPOS"));

   fptr->got_rec = YES;
   return(IOGOOD);
}

/*
 *  Find first record in the file
*/
static int i_firstkey(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];

   if (stat = BTRV(B_GETLOWEST+BT_LOCK, (char *)bptr->fcb, buffer, &fptr->rec_len, bptr->key_buf, fptr->cur_key))
      return((stat == BTNO_MORE_RECS) ? IOEOF : io_xlate(fd_sys, stat, "BTRV FRST"));

   if (stat = BTRV(B_GETPOSN, (char *)bptr->fcb, (char *)&bptr->recposn, &posnlen, NULL, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV FIRSTPOS"));

   fptr->got_rec = YES;
   return(IOGOOD);
}

/*
 *  Find last physical record in the file
*/
static int i_lastkey(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];

   if (stat = BTRV(B_GETLAST+BT_LOCK, (char *)bptr->fcb, buffer, &fptr->rec_len, bptr->key_buf, fptr->cur_key))
      return((stat == BTNO_MORE_RECS) ? IOEOF : io_xlate(fd_sys, stat, "BTRV LAST"));

   if (stat = BTRV(B_GETPOSN, (char *)bptr->fcb, (char *)&bptr->recposn, &posnlen, NULL, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV LASTPOS"));

   fptr->got_rec = YES;
   return(IOGOOD);
}

/*
 *  Find next record in the file
*/
static int i_nextrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];

   if (stat = BTRV(B_GETNEXT+BT_LOCK, (char *)bptr->fcb, buffer, &fptr->rec_len, bptr->key_buf, fptr->cur_key))
      return((stat == BTNO_MORE_RECS) ? IOEOF : io_xlate(fd_sys, stat, "BTRV NEXT"));

   if (stat = BTRV(B_GETPOSN, (char *)bptr->fcb, (char *)&bptr->recposn, &posnlen, NULL, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV NEXTPOS"));

   fptr->got_rec = YES;
   return(IOGOOD);
}

/*
 *  Find previous record in the file
*/
static int i_prevrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];

   if (stat = BTRV(B_GETPREV+BT_LOCK, (char *)bptr->fcb, buffer, &fptr->rec_len, bptr->key_buf, fptr->cur_key))
      return((stat == BTNO_MORE_RECS) ? IOTOF : io_xlate(fd_sys, stat, "BTRV PREV"));

   if (stat = BTRV(B_GETPOSN, (char *)bptr->fcb, (char *)&bptr->recposn, &posnlen, NULL, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV PREVPOS"));

   fptr->got_rec = YES;
   return(IOGOOD);
}

/*
 *  Re-read/reposition record pointer function - if required
*/
static int i_rereadrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   fd[fd_sys].got_rec = NO;
   return(i_findkey(fd_sys, buffer));
}


/*
 *  Add a new record.
*/
static int i_addrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];

   fptr->got_rec = NO;
   fptr->cur_key = 0;

   if (stat = BTRV(B_INSERT, (char *)bptr->fcb, buffer, &fptr->rec_len, bptr->key_buf, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV ADD1"));

   if (stat = BTRV(B_GETPOSN, (char *)bptr->fcb, (char *)&bptr->recposn, &posnlen, NULL, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV ADDPOS"));

   memcpy(buffer, (char *)&bptr->recposn, 4);
   if (stat = BTRV(B_GETDIRECT+BT_LOCK, (char *)bptr->fcb, buffer, &fptr->rec_len, bptr->key_buf, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV ADD2"));

   return(IOGOOD);
}

/*
 *  Update the current record.
*/
static int i_updrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];

   if (stat = BTRV(B_UPDATE, (char *)bptr->fcb, buffer, &fptr->rec_len, bptr->key_buf, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV UPD1"));

   memcpy(buffer, (char *)&bptr->recposn, 4);
   if (stat = BTRV(B_GETDIRECT+BT_LOCK, (char *)bptr->fcb, buffer, &fptr->rec_len, bptr->key_buf, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV UPD2"));

   return(IOGOOD);
}

/*
 *  Delete the current record.
*/
static int i_delrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];

   if (stat = BTRV(B_DELETE, (char *)bptr->fcb, NULL, &fptr->rec_len, NULL, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV DEL"));

   return(IOGOOD);
}

/*
 *  Lock Record
 *  Btrieve locks on its function calls by adding the B_LOCK or B_NWLOCK
 *  value to the operation code (or PRO-C NOLOCK value of 0)
 *  This does a direct get to the physical location of the record to
 *  reposition the record pointer for a write, with a lock
*/  
static int i_lockrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];
   static char locbuf[MAX_RECLEN];    /* stops over-writing new rec on read */

   memcpy(locbuf, (char *)&bptr->recposn, 4);
   if (stat = BTRV(B_GETDIRECT+BT_LOCK, (char *)bptr->fcb, locbuf, &fptr->rec_len, bptr->key_buf, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV LOCKGET"));

   return(IOGOOD);
}

/*
 *  Unlock Record
*/
static int i_unlock_rec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;
   struct fd_def *fptr = &fd[fd_sys];
   struct btr_def *bptr = &fbtr[fd_sys];

   if (stat = BTRV(B_ULOCK, (char *)bptr->fcb, NULL, &fptr->rec_len, NULL, fptr->cur_key))
      return(io_xlate(fd_sys, stat, "BTRV ULCK"));

   return(IOGOOD);
}

/*
 *  Login
*/
static int i_login(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Logoff
*/
static int i_logoff(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  End (Commit) transaction 
*/
static int i_commit(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;

   if (BT_TRANS && (stat = BTRV(B_END_TRN, NULL, NULL, NULL, NULL, 0)))
   {
      errmsg(FileTrnsCommit);
      return(io_xlate(fd_sys, stat, "BTRV ETRN"));
   }
   return(IOGOOD);
}

/*
 *  Rollback transaction
*/
static int i_rollback(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;

   if (BT_TRANS && (stat = BTRV(B_ABT_TRN, NULL, NULL, NULL, NULL, 0)))
   {
      errmsg(FileTrnsRlback);
      return(io_xlate(fd_sys, stat, "BTRV ATRN"));
   }
   return(IOGOOD);
}

/*
 *  Start transaction
 *  - record lock parameter added to Btrieve call
*/
static int i_transact(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int stat;

   if (BT_TRANS && (stat = BTRV(BT_LOCK+B_BEG_TRN, NULL, NULL, NULL, NULL, 0)))
   {
      errmsg(FileTrnsBegin);
      return(io_xlate(fd_sys, stat, "BTRV BTRN"));
   }
   return(IOGOOD);
}

/*
 *  This routine translates BTRIEVE Error codes into PRO-C error codes.
 *  If no PRO-C equivalent, displays the error number.
*/
static int io_xlate(fd_sys, ernum, rtnname)
int fd_sys;
int ernum;
char *rtnname;
{
   switch(ernum)
   {
      case BTBADOPEN   : 
         return(IOBADOPEN);
      case BTDUP       : 
         return(IODUP);
      case BTKEYNFOUND : 
         return(IONOKEY);
      case BTINDEXPOSN : 
         return(IOINDEXPOSN);
      case BTNO_FILE   : 
         return(IONOFILE);
      case BTRECINUSE  :
      case BFILEINUSE  :
         return(IOLOCKED);
      case BTTRANSERR  :
      case BTTRANSACTV :
      case BTTRANSFIO  :
      case BTTRANSEND  :
      case BTTRANSMAX  :
         return(IONOTRANS);
   }

   if (fd_sys >= 0)
      errmsg(FileDbgError_sdss, "Btrieve", ernum, fd[fd_sys].filname, rtnname);
   else      /* if routines called from generated apps without valid fd_sys */
      errmsg(FileDbgError_sds, "Btrieve", ernum, rtnname);
   return(IOERROR);
}


/*
 * Assign section
*/

void assign_IO_BT(dbnum)
int dbnum;
{
   Fntab[dbnum - 1][0]  = (int(*)())0; /* Empty */
   Fntab[dbnum - 1][1]  = i_init_file;
   Fntab[dbnum - 1][2]  = i_open_file;
   Fntab[dbnum - 1][3]  = i_close_file;
   Fntab[dbnum - 1][4]  = i_addrec;
   Fntab[dbnum - 1][5]  = i_delrec;
   Fntab[dbnum - 1][6]  = i_findkey;
   Fntab[dbnum - 1][7]  = i_firstkey;
   Fntab[dbnum - 1][8]  = i_lastkey;
   Fntab[dbnum - 1][9]  = i_lockrec;
   Fntab[dbnum - 1][10] = i_nextrec;
   Fntab[dbnum - 1][11] = i_prevrec;
   Fntab[dbnum - 1][12] = i_unlock_rec;
   Fntab[dbnum - 1][13] = i_updrec;
   Fntab[dbnum - 1][14] = i_commit;
   Fntab[dbnum - 1][15] = i_login;
   Fntab[dbnum - 1][16] = i_logoff;
   Fntab[dbnum - 1][17] = i_rollback;
   Fntab[dbnum - 1][18] = i_transact;
   Fntab[dbnum - 1][19] = i_selectinx;
   Fntab[dbnum - 1][20] = i_rereadrec;
   Fntab[dbnum - 1][21] = i_filename;
}
