/* ==( io/src/iprotree.c )== */
/* ----------------------------------------------- */
/* Pro-C  Copyright (C) 1988 - 1990 Vestronix Inc. */
/* Modification to this source is not supported    */
/* by Vestronix Inc.                               */
/*            All Rights Reserved                  */
/* ----------------------------------------------- */
/* Written   NA   24-Feb-88                        */
/* Modified  BRC  12-Jan-90  See comments below    */
/* ----------------------------------------------- */
/* %W%  (%H% %T%) */

/*
**  Modifications
**
**  12-Feb-90  BRC - IOGEN'd version
**   4-Dec-89  BRC - multi-user, dynamic levels, free space reuse
**     Apr-89  JZ  - Rewrote the indexing functions for single user
**     Dec-88  VvA - Wrote indexing functions for ASCII file manager
**  23-Mar-88  GEO - Changes ASCII file manager
**  24-Feb-88  NIG - Wrote ASCII file manager
*/

/*
**  1.90b - Fixed Length PRO-TREE IOGEN interface
**
**  Features:                                                                *
**     - Dynamic leveling                                                    *
**     - Free list for data file and each key in the index file              *
**     - Single-user or Multi-user with appropriate index file locking       *
**     - Single record locking on the data file                              *
*/
#include "protree.h"


/* Function prototypes */
#ifdef ANSI
int i_lockrec(int, char *);
int i_unlock_rec(int, char *);
static int i_addrec(int, char *);
static int i_close_file(int, char *);
static int i_commit(int, char *);
static int i_delrec(int, char *);
static int i_filename(int, char *);
static int i_findkey(int, char *);
static int i_firstkey(int, char *);
static int i_init_file(int, char *);
static int i_lastkey(int, char *);
static int i_login(int, char *);
static int i_logoff(int, char *);
static int i_nextrec(int, char *);
static int i_open_file(int, char *);
static int i_prevrec(int, char *);
static int i_rereadrec(int, char *);
static int i_rollback(int, char *);
static int i_selectinx(int, char *);
static int i_transact(int, char *);
static int i_updrec(int, char *);
#else
int i_lockrec();
int i_unlock_rec();
static int i_addrec();
static int i_close_file();
static int i_commit();
static int i_delrec();
static int i_filename();
static int i_findkey();
static int i_firstkey();
static int i_init_file();
static int i_lastkey();
static int i_login();
static int i_logoff();
static int i_nextrec();
static int i_open_file();
static int i_prevrec();
static int i_rereadrec();
static int i_rollback();
static int i_selectinx();
static int i_transact();
static int i_updrec();
#endif

/* EXTERNs */
extern long lseek();
extern char *alloc();
extern long lseek();
extern char *alloc();

extern FDDEF fd[]; /* iogen.c */


#ifdef LOCKDEBUG
static FILE *PT_fp;
static char *LogFile = "/tmp/locklog";
#endif


/*
**  Set up File Name extension
*/
static int
i_filename(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   static char *datext = ".dat";
   static char *idxext = ".idx";
   AFDDEF *aptr = &PT_fd[fd_sys];
   int i, iostat = IOGOOD;

   aptr->fname = alloc(strlen(buffer) + strlen(datext) + 1);
   aptr->iname = alloc(strlen(buffer) + strlen(idxext) + 1);
   strcpy(aptr->fname, buffer);
   strcpy(aptr->iname, buffer);

   for (i = strlen(buffer) - 1; i >= 0; i--)
      switch (buffer[i])
      {
      case ']': case ':': case '/': case '\\':
	 i = 0;
	 break;

      case '.':
	 strcpy(aptr->iname + i, idxext);
	 if (strlen(aptr->fname) - 1 == i)
	    strcpy(aptr->fname + i, datext);
	 i = -2;
	 break;
      }
   if (i == -1)
   {
      strcat(aptr->fname, datext);
      strcat(aptr->iname, idxext);
   }

#ifdef LOCKDEBUG
   PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s FILENAME (%s)\r\n", ttyname(0), aptr->fname,
     io_tran(iostat));
   fclose(PT_fp);
#endif

   return(iostat);
}


/*
**  Initialization file descriptors to -1
*/
static int
i_init_file(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int iostat = IOGOOD;

   if (sizeof(WORD8) != 1 || sizeof(WORD16) != 2 || sizeof(WORD32) != 4 ||
     sizeof(FLOAT32) != 4 || sizeof(FLOAT64) != 8)
   {
      errmsg("PRO-TREE setup is incorrect, run PROTEST for details");
      iostat = IOCONFIG;
   }

   if (iostat == IOGOOD && PT_filecnt <= 0)
   {
      PT_fd = (AFDDEF *)alloc(sizeof(AFDDEF) * MAX_FILES);
      PT_filecnt = 0;
      PT_Reclen = 0;
      PT_Nodelen = 0;
   }

#ifdef LOCKDEBUG
   PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: INIT (%s)\r\n", ttyname(0), io_tran(iostat));
   fclose(PT_fp);
#endif

   return(IOGOOD);
}


/*
**  File open function - required parameters are in fd structure
*/
static int
i_open_file(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   FDDEF *fptr = &fd[fd_sys];
   AFDDEF *aptr = &PT_fd[fd_sys];
   int fds, iostat = IOGOOD;

   if (fptr->rec_len > PT_Reclen)
   {
      if (PT_Reclen > 0 && PT_Recbuf)
	 free(PT_Recbuf);
      PT_Recbuf = alloc(PT_Reclen = fptr->rec_len);
   }

   /* set up values */
   memcpy(aptr->version, CURRENT_VERSION, 3);
   aptr->maxelts = MAXELTS;
   aptr->ilocked = 0;
   aptr->doffset = PTNOPTR;

#ifdef LOCKDEBUG
   PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s OPEN top\r\n", ttyname(0), aptr->fname);
   fclose(PT_fp);
#endif

   /* Open up the INDEX file */
   if ((fds = OPN_FILE(aptr->iname)) < 0)
   {
      if ((fds = CRT_FILE(aptr->iname)) >= 0)
      {
	 aptr->fd_idx = fds;

	 if (PT_idxlock(fd_sys, FILE_LOCK) != IOGOOD)
	 {
#ifdef MSDOS
	    errmsg("ERROR: File Locking Error or SHARE Not Running\n");
#else
	    errmsg("ERROR: File Locking Error: race condition during create\n");
#endif
	    iostat = IOFILLOCK;
	 }

	 if (iostat == IOGOOD)
	 {
	    if ((iostat = PT_writehdr(fd_sys)) != IOGOOD)
	    {
	       /* delete the corrupt index file */
	       aptr->fd_idx = -1;
	       close(fds);
	       unlink(aptr->iname);
	    }

	    /* delete any associated orphaned data file */
	    unlink(aptr->fname);
	    if ((fds = CRT_FILE(aptr->fname)) < 0)
	       iostat = IODATPERM;
	    else
	       fptr->fd_num = fds;
	 }

#ifdef MULTI_USER
	 if (iostat == IOGOOD)
	    iostat = PT_idxunlock(fd_sys, FILE_LOCK);
#endif
      }
      else
	 iostat = IOIDXPERM;
   }
   else
   {
      aptr->fd_idx = fds;

#ifdef MULTI_USER
      iostat = PT_idxlock(fd_sys, READ_LOCK);
#endif

      if (iostat == IOGOOD)
	 iostat = PT_readhdr(fd_sys);

      /* Open up the DATA file */
      if (iostat == IOGOOD)
      {
	 if ((fds = OPN_FILE(aptr->fname)) < 0)
	    iostat = IODATPERM;
	 else
	    fptr->fd_num = fds;
      }

#ifdef MULTI_USER
      PT_idxunlock(fd_sys, READ_LOCK);
#endif
   }

   if (iostat != IOGOOD)
      i_close_file(fd_sys, buffer);
   else
   {
      if (fptr->rec_len > PT_Reclen)
      {
	 if (PT_Reclen > 0 && PT_Recbuf)
	    free(PT_Recbuf);
	 PT_Recbuf = alloc(PT_Reclen = fptr->rec_len);
      }

      aptr->doffset = PTNOPTR;
      PT_filecnt++;
   }

#ifdef LOCKDEBUG
   PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s OPEN: fd_num=%d\r\n", ttyname(0), aptr->fname,
	  fptr->fd_num);
   fprintf(PT_fp, "%s: %s OPEN: fd_idx=%d\r\n", ttyname(0), aptr->iname,
	  aptr->fd_idx);
   fprintf(PT_fp, "%s: %s OPEN end (%s)\r\n", ttyname(0), aptr->fname,
     io_tran(iostat));
   fclose(PT_fp);
#endif

   return(iostat);
}


/*
**  Close the specified file
*/
static int
i_close_file(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   FDDEF *fptr;
   AFDDEF *aptr;
   int ix, iostat;

   if ((iostat = PT_chk_open(fd_sys, "CLOSE")) != IOGOOD)
      return(iostat);

   fptr = &fd[fd_sys];
   aptr = &PT_fd[fd_sys];

#ifdef LOCKDEBUG
      PT_fp = fopen(LogFile, "a");
      fprintf(PT_fp, "%s: %s CLOSE top\r\n", ttyname(0), aptr->fname);
      fclose(PT_fp);
#endif

#ifndef MULTI_USER
   if (aptr->fd_idx >= 0)
   {
      if ((iostat = PT_flush_cache(fd_sys)) == IOGOOD)
	 iostat = PT_idxunlock(fd_sys, FILE_LOCK);
      else
	 PT_idxunlock(fd_sys, FILE_LOCK);
   }
#endif

   if (aptr->fd_idx >= 0)
      close(aptr->fd_idx);      /* close the index file */

   if (fptr->fd_num >= 0)
      close(fptr->fd_num);      /* close the data file */

   if (aptr->fd_idx >= 0)
   {
#ifdef LOCKDEBUG
      PT_fp = fopen(LogFile, "a");
      fprintf(PT_fp, "%s: %s CLOSE mid (%s)\r\n", ttyname(0), aptr->fname,
        io_tran(iostat));
      fclose(PT_fp);
#endif
      if (aptr->keyinfo)
      {
	 for (ix = 0; ix < fptr->key_cnt; ix++)
	    PT_freenodes(fd_sys, &(aptr->keyinfo[ix]));

	 free((char *)aptr->keyinfo);
	 aptr->keyinfo = NULL;
      }
      if (aptr->iname)
      {
	 free(aptr->iname);
	 aptr->iname = NULL;
      }
      if (aptr->fname)
      {
	 free(aptr->fname);
	 aptr->fname = NULL;
      }
   }

   fptr->fd_num = -1;
   aptr->fd_idx = -1;

   if (--PT_filecnt <= 0)
   {
      if (PT_fd)
	 free((char *)PT_fd);
      if (PT_Reclen > 0 && PT_Recbuf)
	 free(PT_Recbuf);
      if (PT_Nodelen > 0 && PT_Nodebuf)
	 free(PT_Nodebuf);
      PT_filecnt = -1;
   }

#ifdef LOCKDEBUG
      PT_fp = fopen(LogFile, "a");
      fprintf(PT_fp, "%s: %s CLOSE end (%s)\r\n", ttyname(0), fptr->filname,
        io_tran(iostat));
      fclose(PT_fp);
#endif

   return(iostat);
}


/*
**  Select an index to perform processing on - done in IOGEN.C
*/
static int
i_selectinx(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
**  Find and lock a record by an EXACT or PARTIAL key value
*/
static int
i_findkey(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   FDDEF *fptr;
   AFDDEF *aptr;
   KINFO *kptr;
   int iostat, lev, elt;

   if ((iostat = PT_chk_open(fd_sys, "FIND")) != IOGOOD)
      return(iostat);

   fptr = &fd[fd_sys];
   aptr = &PT_fd[fd_sys];
   kptr = &(aptr->keyinfo[fptr->cur_key]);

#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] FIND top\r\n", ttyname(0), aptr->fname,
     fptr->cur_key);
   /*fclose(PT_fp);*/
#endif

   /* prepare a buffer and extract key from record into it */
   if (!kptr->keyval)
      kptr->keyval = alloc(kptr->k_leng);
   PT_get_key(fd_sys, buffer, kptr->keyval);

#ifdef MULTI_USER
   /* apply for a READ lock */
   iostat = PT_idxlock(fd_sys, READ_LOCK);
#endif

   if (iostat == IOGOOD)
      iostat = i_unlock_rec(fd_sys, buffer);
   else
      i_unlock_rec(fd_sys, buffer);

   if (iostat == IOGOOD)
   {
      if (!fptr->exact)
      {
	 aptr->doffset = PT_gteq_key(fd_sys, kptr->keyval, FALSE);
      }
      else
      {
	 /* get key >= to allow for Duplicate suffix */
	 aptr->doffset = PT_gteq_key(fd_sys, kptr->keyval,
	   kptr->k_type == KEY_UNIQUE);

	 if (aptr->doffset != PTNOPTR && kptr->k_type == KEY_DUPLICATE)
	 {
	    lev = kptr->nodebuff[0].level;
	    elt = kptr->c_elt[lev];
	    if (PT_keycmp(fd_sys, kptr->keyval,
	      kptr->nodebuff[lev].elt[elt].keyval, KEY_UNIQUE))
	       aptr->doffset = PTNOPTR;
	 }
      }
      if (aptr->doffset == PTNOPTR)
	 iostat = fptr->exact ? IONOKEY : IOEOF;
   }

   if (iostat == IOGOOD)
   {
      iostat = PT_get_rec(fd_sys, buffer);
   }
   else if (kptr->keyval)
   {
      free(kptr->keyval);
      kptr->keyval = NULL;
   }

#ifdef MULTI_USER
   /* release the READ lock */
   if (iostat == IOGOOD)
      iostat = PT_idxunlock(fd_sys, READ_LOCK);
   else
      PT_idxunlock(fd_sys, READ_LOCK);
#endif

#ifdef LOCKDEBUG
   /*PT_fp = fopen(LogFile, "a");*/
   fprintf(PT_fp, "%s: %s[%d] FIND end (%s)\r\n", ttyname(0), aptr->fname,
     fptr->cur_key, io_tran(iostat));
   fclose(PT_fp);
#endif

   return(iostat);
}


/*
**  Find the first record in the specified file
*/
static int
i_firstkey(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   FDDEF *fptr;
   AFDDEF *aptr;
   KINFO *kptr;
   int iostat;

   if ((iostat = PT_chk_open(fd_sys, "FIRST")) != IOGOOD)
      return(iostat);

   fptr = &fd[fd_sys];
   aptr = &PT_fd[fd_sys];
   kptr = &(aptr->keyinfo[fptr->cur_key]);

#ifdef LOCKDEBUG
   PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] FIRST top\r\n", ttyname(0), aptr->fname,
     fptr->cur_key);
   fclose(PT_fp);
#endif

#ifdef MULTI_USER
   /* apply for a READ lock */
   iostat = PT_idxlock(fd_sys, READ_LOCK);
#endif

   if (iostat == IOGOOD)
      iostat = i_unlock_rec(fd_sys, buffer);
   else
      i_unlock_rec(fd_sys, buffer);

   /* retrieve current data record offset */
   if (iostat == IOGOOD)
      if ((aptr->doffset = PT_first_key(fd_sys)) == PTNOPTR)
			iostat = IOEOF;

   if (iostat == IOGOOD && aptr->doffset != PTNOPTR)
      iostat = PT_get_rec(fd_sys, buffer);
   else if (kptr->keyval)
   {
      free(kptr->keyval);
      kptr->keyval = NULL;
   }

#ifdef MULTI_USER
   /* release the READ lock */
   if (iostat == IOGOOD)
      iostat = PT_idxunlock(fd_sys, READ_LOCK);
   else
      PT_idxunlock(fd_sys, READ_LOCK);
#endif

#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] FIRST end (%s)\r\n", ttyname(0),
     aptr->fname, fptr->cur_key, io_tran(iostat));
   fclose(PT_fp);
#endif

   return(iostat);
}


/*
**  Find the last record in the specified file
*/
static int
i_lastkey(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   FDDEF *fptr;
   AFDDEF *aptr;
   KINFO *kptr;
   int iostat;

   if ((iostat = PT_chk_open(fd_sys, "LAST")) != IOGOOD)
      return(iostat);

   fptr = &fd[fd_sys];
   aptr = &PT_fd[fd_sys];
   kptr = &(aptr->keyinfo[fptr->cur_key]);

#ifdef LOCKDEBUG
   PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] LAST top\r\n", ttyname(0), aptr->fname,
     fptr->cur_key);
   fclose(PT_fp);
#endif

#ifdef MULTI_USER
   /* apply for a READ lock */
   iostat = PT_idxlock(fd_sys, READ_LOCK);
#endif

   if (iostat == IOGOOD)
      iostat = i_unlock_rec(fd_sys, buffer);
   else
      i_unlock_rec(fd_sys, buffer);

   /* retrieve current data record offset */
   if (iostat == IOGOOD)
      if ((aptr->doffset = PT_last_key(fd_sys)) == PTNOPTR)
	 iostat = IOEOF;

   if (iostat == IOGOOD)
      iostat = PT_get_rec(fd_sys, buffer);
   else if (kptr->keyval)
   {
      free(kptr->keyval);
      kptr->keyval = NULL;
   }

#ifdef MULTI_USER
   /* release the READ lock */
   if (iostat == IOGOOD)
      iostat = PT_idxunlock(fd_sys, READ_LOCK);
   else
      PT_idxunlock(fd_sys, READ_LOCK);
#endif

#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] LAST end (%s)\r\n", ttyname(0),
     aptr->fname, fptr->cur_key, io_tran(iostat));
   fclose(PT_fp);
#endif

   return(iostat);
}


/*
**  Find the next record in the specified file
**
**  Although IOGEN does checking for duplicate key records in an EXACT
**  search, ASCII sequential files are not indexed; thus this function
**  will continue traversing the file to find all occurrences matching
**  the original exact-matched key.
*/
static int
i_nextrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   FDDEF *fptr;
   AFDDEF *aptr;
   KINFO *kptr;
   int iostat, lev, elt;

   if ((iostat = PT_chk_open(fd_sys, "NEXT")) != IOGOOD)
      return(iostat);

   fptr = &fd[fd_sys];
   aptr = &PT_fd[fd_sys];
   kptr = &(aptr->keyinfo[fptr->cur_key]);

#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] NEXT top\r\n", ttyname(0), aptr->fname,
     fptr->cur_key);
   fclose(PT_fp);
#endif

#ifdef MULTI_USER
   /* apply for a READ lock */
   iostat = PT_idxlock(fd_sys, READ_LOCK);
#endif

   if (iostat == IOGOOD && !kptr->keyval)
      iostat = IONOKEY;
   else if (iostat == IOGOOD)
      iostat = i_unlock_rec(fd_sys, buffer);
   else
      i_unlock_rec(fd_sys, buffer);

   /* retrieve current data record offset */
   if (iostat == IOGOOD)
   {
      long offset;
      if (PT_gteq_key(fd_sys, kptr->keyval, TRUE) != PTNOPTR)
	 offset = PT_next_key(fd_sys);
      else
	 offset = PT_gteq_key(fd_sys, kptr->keyval, FALSE);
      if (offset == PTNOPTR)
	 iostat = IOEOF;
      else
      {
	 aptr->doffset = offset;
	 if (fptr->exact)
	 {
	    lev = kptr->nodebuff[0].level;
	    elt = kptr->c_elt[lev];
	    if (PT_keycmp(fd_sys, kptr->keyval,
	      kptr->nodebuff[lev].elt[elt].keyval, KEY_UNIQUE))
	    {
	       iostat = IONONEXT;
	       aptr->doffset = PTNOPTR;
	    }
	 }
      }
   }

   if (iostat == IOGOOD)
      iostat = PT_get_rec(fd_sys, buffer);

#ifdef MULTI_USER
   /* release the READ lock */
   if (iostat == IOGOOD)
      iostat = PT_idxunlock(fd_sys, READ_LOCK);
   else
      PT_idxunlock(fd_sys, READ_LOCK);
#endif

#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] NEXT end (%s)\r\n", ttyname(0),
     aptr->fname, fptr->cur_key, io_tran(iostat));
   fclose(PT_fp);
#endif

   return(iostat);
}


/*
**  Find the previous record in the specified file
**
**  Although IOGEN does checking for duplicate key records in an EXACT
**  search, ASCII sequential files are not indexed; thus this function
**  will continue traversing the file to find all occurrences matching
**  the original exact-matched key.
*/
static int
i_prevrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   FDDEF *fptr;
   AFDDEF *aptr;
   KINFO *kptr;
   int iostat, lev, elt;

   if ((iostat = PT_chk_open(fd_sys, "PREV")) != IOGOOD)
      return(iostat);

   fptr = &fd[fd_sys];
   aptr = &PT_fd[fd_sys];
   kptr = &(aptr->keyinfo[fptr->cur_key]);

#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] PREV top\r\n", ttyname(0), aptr->fname,
     fptr->cur_key);
   fclose(PT_fp);
#endif

#ifdef MULTI_USER
   /* apply for a READ lock */
   iostat = PT_idxlock(fd_sys, READ_LOCK);
#endif

   if (iostat == IOGOOD && !kptr->keyval)
      iostat = IONOKEY;
   else if (iostat == IOGOOD)
      iostat = i_unlock_rec(fd_sys, buffer);
   else
      i_unlock_rec(fd_sys, buffer);

   /* retrieve current data record offset */
   if (iostat == IOGOOD)
   {
      long offset;

      offset = PT_gteq_key(fd_sys, kptr->keyval, TRUE);
      offset = PT_prev_key(fd_sys);
      if (offset == PTNOPTR)
	 iostat = IOTOF;
      else
      {
	 aptr->doffset = offset;
	 if (fptr->exact)
	 {
	    lev = kptr->nodebuff[0].level;
	    elt = kptr->c_elt[lev];
	    if (PT_keycmp(fd_sys, kptr->keyval,
	      kptr->nodebuff[lev].elt[elt].keyval, KEY_UNIQUE))
	    {
	       iostat = IONONEXT;
	       aptr->doffset = PTNOPTR;
	    }
	 }
      }
   }

   if (iostat == IOGOOD)
      iostat = PT_get_rec(fd_sys, buffer);

#ifdef MULTI_USER
   /* release the READ lock */
   if (iostat == IOGOOD)
      iostat = PT_idxunlock(fd_sys, READ_LOCK);
   else
      PT_idxunlock(fd_sys, READ_LOCK);
#endif

#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] PREV end (%s)\r\n", ttyname(0),
     aptr->fname, fptr->cur_key, io_tran(iostat));
   fclose(PT_fp);
#endif

   return(iostat);
}


/*
**  re-read/reposition record pointer function - dummy function here
*/
static int
i_rereadrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
**  ADD the given record to the specified file
*/
static int
i_addrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   FDDEF *fptr;
   AFDDEF *aptr;
   KINFO *kptr;
   int ck, ix, iostat;

   if ((iostat = PT_chk_open(fd_sys, "ADD")) != IOGOOD)
      return(iostat);

   fptr = &fd[fd_sys];
   aptr = &PT_fd[fd_sys];

   /* save current index */
   ck = fptr->cur_key;

#ifdef LOCKDEBUG
   PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] ADD top\r\n", ttyname(0), aptr->fname,
     fptr->cur_key);
   fclose(PT_fp);
#endif

#ifdef MULTI_USER
   /* apply for a WRITE lock */
   iostat = PT_idxlock(fd_sys, WRITE_LOCK);
#endif

   if (iostat == IOGOOD)
      iostat = i_unlock_rec(fd_sys, buffer);
   else
      i_unlock_rec(fd_sys, buffer);

   /* pre-add check for all unique keys */
   for (ix = 0; iostat == IOGOOD && ix < fptr->key_cnt; ix++)
   {
      fptr->cur_key = ix;
      kptr = &(aptr->keyinfo[ix]);
      if (!kptr->keyval)
	 kptr->keyval = alloc(kptr->k_leng);
      PT_get_key(fd_sys, buffer, kptr->keyval);

      if (kptr->k_type == KEY_UNIQUE)
	 if (PT_gteq_key(fd_sys, kptr->keyval, TRUE) != PTNOPTR)
	    iostat = IODUP;
   }

   /* write the record to the data file at a new offset */
   if (iostat == IOGOOD)
   {
      aptr->doffset = PTNOPTR;
      iostat = PT_writedata(fd_sys, buffer);
   }

   /* update each index */
   for (ix = 0; iostat == IOGOOD && ix < fptr->key_cnt; ix++)
   {
      fptr->cur_key = ix;
      kptr = &(aptr->keyinfo[ix]);
      if (kptr->k_type == KEY_DUPLICATE)
	 iostat = PT_new_dup(fd_sys, kptr->keyval);
      if (iostat == IOGOOD)
	 iostat = PT_add_key(fd_sys, kptr->keyval);
   }

   /* release ALL keyvals after the add */
   for (ix = 0; ix < fptr->key_cnt; ix++)
   {
      kptr = &(aptr->keyinfo[ix]);
      if (ix != ck && kptr->keyval)
      {
	 free(kptr->keyval);
	 kptr->keyval = NULL;
      }
   }

#ifdef MULTI_USER
   /* release the WRITE lock */
   if (iostat == IOGOOD)
      iostat = PT_idxunlock(fd_sys, WRITE_LOCK);
   else
      PT_idxunlock(fd_sys, WRITE_LOCK);
#endif

   /* restore current index */
   fptr->cur_key = ck;

   if (iostat == IOGOOD)
      iostat = i_unlock_rec(fd_sys, buffer);
   else
      i_unlock_rec(fd_sys, buffer);

#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] ADD end (%s)\r\n", ttyname(0),
     aptr->fname, fptr->cur_key, io_tran(iostat));
   fclose(PT_fp);
#endif

   return(iostat);
}


/*
**  UPDATE the current record in the specified file
*/
static int
i_updrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   FDDEF *fptr;
   AFDDEF *aptr;
   KINFO *kptr;
   char *keyval, *kval;
   int iostat, ck, ix;
   long offset;

   if ((iostat = PT_chk_open(fd_sys, "UPDATE")) != IOGOOD)
      return(iostat);

   fptr = &fd[fd_sys];
   aptr = &PT_fd[fd_sys];
   kptr = &(aptr->keyinfo[fptr->cur_key]);
   kval = kptr->keyval;

   /* save the current key */
   ck = fptr->cur_key;

#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] UPDATE top\r\n", ttyname(0), aptr->fname,
     fptr->cur_key);
   fclose(PT_fp);
#endif

#ifdef MULTI_USER
   /* apply for a WRITE lock */
   iostat = PT_idxlock(fd_sys, WRITE_LOCK);
#endif

   /* make sure that the record is still there and locked */
   if (iostat == IOGOOD && (!kptr->keyval || aptr->doffset == PTNOPTR))
      iostat = IONOKEY;

   /* make sure that the record is still there */
   if (iostat == IOGOOD)
   {
      offset = PT_gteq_key(fd_sys, kptr->keyval, TRUE);
      if (offset == PTNOPTR)
	 iostat = IOERROR;
      else if (aptr->doffset != offset)
	 abort_mess("UPDREC: fptr->offset=%ld, offset=%ld", aptr->doffset,
	   offset);
      else
	 iostat = PT_readdata(fd_sys, PT_Recbuf);
   }

   /* check that all key updates will succeed */
   for (ix = 0; iostat == IOGOOD && ix < fptr->key_cnt; ix++)
   {
      fptr->cur_key = ix;
      kptr = &(aptr->keyinfo[ix]);

      /* grab the key info */
      if (ix != ck)
	 iostat = PT_find_key(fd_sys, PT_Recbuf);

      /* check to see if the key value has changed and if it is ok */
      keyval = alloc(kptr->k_leng);
      if (iostat == IOGOOD)
      {
	 PT_get_key(fd_sys, buffer, keyval);
	 if (!PT_keycmp(fd_sys, keyval, kptr->keyval, KEY_UNIQUE))
	 {
	    if (ix != ck)
	       free(kptr->keyval);
	    kptr->keyval = NULL;
	 }
	 else
	 {
	    if (kptr->k_type == KEY_UNIQUE &&
	      PT_gteq_key(fd_sys, keyval, TRUE) != PTNOPTR)
	       iostat = IODUP;
	 }
      }
      free(keyval);
   }

   /* the update is valid so actually update the indicies now */
   for (ix = 0; iostat == IOGOOD && ix < fptr->key_cnt; ix++)
   {
      fptr->cur_key = ix;
      kptr = &(aptr->keyinfo[ix]);
      if (kptr->keyval)
      {
	 iostat = PT_del_key(fd_sys, kptr->keyval);
	 if (iostat == IOGOOD)
	    PT_get_key(fd_sys, buffer, kptr->keyval);
	 if (iostat == IOGOOD && kptr->k_type == KEY_DUPLICATE)
	    iostat = PT_new_dup(fd_sys, kptr->keyval);
	 if (iostat == IOGOOD)
	    iostat = PT_add_key(fd_sys, kptr->keyval);
      }
   }

   /* release all keyvals after the add */
   for (ix = 0; ix < fptr->key_cnt; ix++)
   {
      kptr = &(aptr->keyinfo[ix]);
      if (kptr->keyval)
      {
	 if (ix != ck)
	    free(kptr->keyval);
	 kptr->keyval = NULL;
      }
   }

   /* release the WRITE lock */
#ifdef MULTI_USER
   if (iostat == IOGOOD)
      iostat = PT_idxunlock(fd_sys, WRITE_LOCK);
   else
      PT_idxunlock(fd_sys, WRITE_LOCK);
#endif

   /* restore current index */
   fptr->cur_key = ck;
   aptr->keyinfo[ck].keyval = kval;

   if (iostat == IOGOOD)
      iostat = PT_writedata(fd_sys, buffer);
   else
      errmsg("writedata NOT done: %s", io_tran(iostat));

   if (iostat == IOGOOD)
      iostat = i_unlock_rec(fd_sys, buffer);
   else
   {
      errmsg("unlocking despite %s", io_tran(iostat));
      errmsg("unlock now returns %s", io_tran(i_unlock_rec(fd_sys, buffer)));
   }

#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] UPDATE end (%s)\r\n", ttyname(0),
   aptr->fname, fptr->cur_key, io_tran(iostat));
   fclose(PT_fp);
#endif

   return(iostat);
}


/*
**  DELETE the current record in the specified file
*/
static int
i_delrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   FDDEF *fptr;
   AFDDEF *aptr;
   KINFO *kptr;
   int iostat, ck, ix;
   long offset;

   if ((iostat = PT_chk_open(fd_sys, "DELETE")) != IOGOOD)
      return(iostat);

   fptr = &fd[fd_sys];
   aptr = &PT_fd[fd_sys];
   kptr = &(aptr->keyinfo[fptr->cur_key]);

   /* save the current key */
   ck = fptr->cur_key;

#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] DELETE top\r\n", ttyname(0), aptr->fname,
     fptr->cur_key);
   fclose(PT_fp);
#endif

#ifdef MULTI_USER
   /* apply for a WRITE lock */
   iostat = PT_idxlock(fd_sys, WRITE_LOCK);
#endif

   /* make sure that the record is still there and locked */
   if (iostat == IOGOOD && (!kptr->keyval || aptr->doffset == PTNOPTR))
      iostat = IONOKEY;

   /* make sure that the record is still there */
   if (iostat == IOGOOD)
   {
      offset = PT_gteq_key(fd_sys, kptr->keyval, TRUE);
      if (offset == PTNOPTR)
	 iostat = IOERROR;
      else if (aptr->doffset != offset)
	 abort_mess("DELREC: fptr->offset=%ld, offset=%ld", aptr->doffset,
	   offset);
      else
	 iostat = PT_readdata(fd_sys, PT_Recbuf);
   }

   /* check that all appropriate indicies are present */
   for (ix = 0; iostat == IOGOOD && ix < fptr->key_cnt; ix++)
   {
      fptr->cur_key = ix;
      kptr = &(aptr->keyinfo[ix]);

      /* grab the key info */
      if (ix != ck)
	 iostat = PT_find_key(fd_sys, PT_Recbuf);
   }

   /* the update is valid so actually delete the indicies now */
   for (ix = 0; iostat == IOGOOD && ix < fptr->key_cnt; ix++)
   {
      fptr->cur_key = ix;
      kptr = &(aptr->keyinfo[ix]);
      iostat = PT_del_key(fd_sys, kptr->keyval);
   }

   /* release all keyvals after the add */
   for (ix = 0; ix < fptr->key_cnt; ix++)
   {
      kptr = &(aptr->keyinfo[ix]);
      if (kptr->keyval && ix != ck)
      {
	 free(kptr->keyval);
	 kptr->keyval = NULL;
      }
   }

#ifdef MULTI_USER
   /* release the WRITE lock */
   if (iostat == IOGOOD)
      iostat = PT_idxunlock(fd_sys, WRITE_LOCK);
   else
      PT_idxunlock(fd_sys, WRITE_LOCK);
#endif

   /* restore current index */
   fptr->cur_key = ck;

   if (iostat == IOGOOD)
   {
      if ((iostat = PT_recrecycle(fd_sys, aptr->doffset)) == IOGOOD)
	 iostat = i_unlock_rec(fd_sys, buffer);
      else
	 i_unlock_rec(fd_sys, buffer);
   }
   else
      i_unlock_rec(fd_sys, buffer);

#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
   fprintf(PT_fp, "%s: %s[%d] DELETE end (%s)\r\n", ttyname(0),
     aptr->fname, fptr->cur_key, io_tran(iostat));
   fclose(PT_fp);
#endif

   return(iostat);
}


/*
** LOCK the current record in the specified file
**
** NOTE: undefine USELOCK on non-multiuser systems (like non-networked DOS)
*/
int
i_lockrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int iostat = IOGOOD;

#if defined(MULTI_USER) && MAXLOCKS > 0
   FDDEF *fptr;
   AFDDEF *aptr;

   if ((iostat = PT_chk_open(fd_sys, "LOCK")) != IOGOOD)
      return(iostat);

   fptr = &fd[fd_sys];
   aptr = &PT_fd[fd_sys];

   /* find the spot ... */
   if (lseek(fptr->fd_num, aptr->doffset, FDIR_BEGIN) != aptr->doffset)
      iostat = IODATSEEK;

   /* ... and lock it if available. */
   else if (LOCK_TST(fptr->fd_num, aptr->doffset, (long)fptr->rec_len))
   {
#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
      fprintf(PT_fp, "%s: %s[%d] LOCK %ld:%ld FAIL! (errno=%d)\r\n",
        ttyname(0), aptr->fname, fptr->cur_key, aptr->doffset,
        (long)fptr->rec_len, errno);
      fclose(PT_fp);
#endif
      iostat = IOLOCKED;
   }
#ifdef LOCKDEBUG
   else
   {
	PT_fp = fopen(LogFile, "a");
      fprintf(PT_fp, "%s: %s[%d] LOCK %ld:%ld\r\n", ttyname(0), 
        aptr->fname, fptr->cur_key, aptr->doffset, (long)fptr->rec_len);
      fclose(PT_fp);
   }
#endif
#endif

   return(iostat);
}


/*
** UNLOCK the currently LOCKed record in the specified file
**
** NOTE: undefine USELOCK on non-multiuser systems (like non-networked DOS)
*/
int
i_unlock_rec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int iostat = IOGOOD;
#if defined(MULTI_USER) && MAXLOCKS > 0
   FDDEF *fptr;
   AFDDEF *aptr;

   if ((iostat = PT_chk_open(fd_sys, "UNLOCK")) != IOGOOD)
   return(iostat);

   fptr = &fd[fd_sys];
   aptr = &PT_fd[fd_sys];

   if (aptr->doffset != PTNOPTR)
   {
      /* find the spot ... */
      if (lseek(fptr->fd_num, aptr->doffset, FDIR_BEGIN) != aptr->doffset)
	 iostat = IODATSEEK;

   /* ... and unlock it. */
      else if (LOCK_REL(fptr->fd_num, aptr->doffset, (long)fptr->rec_len))
      {
#ifdef LOCKDEBUG
	PT_fp = fopen(LogFile, "a");
	 fprintf(PT_fp, "%s: %s[%d] UNLOCK %ld:%ld FAIL (errno=%d)!\r\n",
	   ttyname(0), aptr->fname, fptr->cur_key, aptr->doffset,
	   (long)fptr->rec_len, errno);
	 fclose(PT_fp);
#endif
	 iostat = IONOLOCK;
      }
#ifdef LOCKDEBUG
      else
      {
	PT_fp = fopen(LogFile, "a");
	 fprintf(PT_fp, "%s: %s[%d] UNLOCK %ld:%ld\r\n", ttyname(0),
	   aptr->fname, fptr->cur_key, aptr->doffset, (long)fptr->rec_len);
	 fclose(PT_fp);
      }
#endif

      /* ... reset the lock pointer. */
      aptr->doffset = PTNOPTR;
   }
#endif /* MULTI_USER */

   return(iostat);
}


/*
**  Login - dummy call
*/
static int
i_login(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
**  Logoff - dummy call
*/
static int
i_logoff(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
**  Start transaction
*/
static int
i_transact(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
**  End (Commit) transaction
*/
static int
i_commit(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
**  Rollback transaction
*/
static int
i_rollback(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
**  Assign section
*/
void
assign_IO_PT(dbnum)
int dbnum;
{
   Fntab[dbnum - 1][0]  = (int(*)())0; /* Empty */
   Fntab[dbnum - 1][1]  = i_init_file;
   Fntab[dbnum - 1][2]  = i_open_file;
   Fntab[dbnum - 1][3]  = i_close_file;
   Fntab[dbnum - 1][4]  = i_addrec;
   Fntab[dbnum - 1][5]  = i_delrec;
   Fntab[dbnum - 1][6]  = i_findkey;
   Fntab[dbnum - 1][7]  = i_firstkey;
   Fntab[dbnum - 1][8]  = i_lastkey;
   Fntab[dbnum - 1][9]  = i_lockrec;
   Fntab[dbnum - 1][10] = i_nextrec;
   Fntab[dbnum - 1][11] = i_prevrec;
   Fntab[dbnum - 1][12] = i_unlock_rec;
   Fntab[dbnum - 1][13] = i_updrec;
   Fntab[dbnum - 1][14] = i_commit;
   Fntab[dbnum - 1][15] = i_login;
   Fntab[dbnum - 1][16] = i_logoff;
   Fntab[dbnum - 1][17] = i_rollback;
   Fntab[dbnum - 1][18] = i_transact;
   Fntab[dbnum - 1][19] = i_selectinx;
   Fntab[dbnum - 1][20] = i_rereadrec;
   Fntab[dbnum - 1][21] = i_filename;
}
