/* ==( io/src/irbase.c )== */

/* ----------------------------------------------- */
/* Pro-C  Copyright (C) 1988 - 1990 Vestronix Inc. */
/* Modification to this source is not supported    */
/* by Vestronix Inc.                               */
/*            All Rights Reserved                  */
/* ----------------------------------------------- */
/* Written   NA   24-Feb-88                        */
/* Modified  VV   15-Dec-89  V.2 modifications     */
/* ----------------------------------------------- */
/* %W%  (%H% %T%) */

/*
 *  Modifications
 *
 *  01-Sep-89  VV  Adapted for R:BASE in general I/O form
*/

/*
 *  -- D R A F T --  R:BASE specific I/O calls via general IOGEN.C interface 
*/


#define LINT_ARGS                           /* to enable prototypes in pi.h */

# include <pi.h>                                      /* R:BASE header file */
# include <iodef.h>
# include <iomsg.h>
# include <proc.io>
# include <bench.h>
# include <benchmsg.h>
# include <iosup.h>


/* Function prototypes */
# ifdef ANSI
static int i_addrec(int, char *);
static int i_close_file(int, char *);
static int i_commit(int, char *);
static int i_delrec(int, char *);
static int i_filename(int, char *);
static int i_findkey(int, char *);
static int i_firstkey(int, char *);
static int i_init_file(int, char *);
static int i_lastkey(int, char *);
static int i_lockrec(int, char *);
static int i_login(int, char *);
static int i_logoff(int, char *);
static int i_nextrec(int, char *);
static int i_open_file(int, char *);
static int i_prevrec(int, char *);
static int i_rereadrec(int, char *);
static int i_rollback(int, char *);
static int i_selectinx(int, char *);
static int i_transact(int, char *);
static int i_unlock_rec(int, char *);
static int i_updrec(int, char *);
static int length_check(int);
static int get_record(int, char *);
static int io_xlate(int, int, char *);
static void setupkey(int, char *, char *);
static void buff2rb(int, char *, char *);
# else
static int i_addrec();
static int i_close_file();
static int i_commit();
static int i_delrec();
static int i_filename();
static int i_findkey();
static int i_firstkey();
static int i_init_file();
static int i_lastkey();
static int i_lockrec();
static int i_login();
static int i_logoff();
static int i_nextrec();
static int i_open_file();
static int i_prevrec();
static int i_rereadrec();
static int i_rollback();
static int i_selectinx();
static int i_transact();
static int i_unlock_rec();
static int i_updrec();
static int length_check();
static int get_record();
static int io_xlate();
static void setupkey();
static void buff2rb();
# endif


extern struct fd_def fd[]; /* iogen.c */

extern int Rberrcod;                /* R:BASE error code (declared in pi.h) */
extern char dbfilename[];         /* CHECK where its master declaration is! */


/*
 * structure for R:BASE-specific values 
*/
struct rbf_def
{
   RELATION *fdptr;         /* R:BASE table pointer to identify this "file" */
   int rbrec_len;                        /* record length as seen by R:BASE */
   int fld_cnt;                                              /* field count */
};

static struct rbf_def frbf[MAX_FILES];

static DATABASE *dbptr;                  /* R:BASE overall database pointer */
static char keybuff[MAX_FLDLEN * MAX_SEGS + MAX_SEGS];
static char storbuff[MAX_RECLEN];
static char wherebuff[255];        /* buffer for "where" clauses for search */


/*
 *
 * Interface Functions
 *
*/


/*
 *  Sets up File Name extension.
*/
static int i_filename(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   for ( ; (*buffer != '.') && (*buffer != NULL); buffer++)
      ;
   *buffer = NULL;

   return(IOGOOD);
}

/*
 *  File initialization - starts up R:BASE engine and opens database
*/
static int i_init_file(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   rbstart();                         /* R:BASE startup, required only once */
   dbptr = rbopen(dbfilename);           /* open overall database once only */
   if (Rberrcod)
      return(io_xlate(fd_sys, Rberrcod, "RB INIT"));

   return(IOGOOD);
}

/*
 *  File open function
 *  - note that R:BASE Program Interface lacks the capability to create an
 *    R:BASE file; this must be done manually from within R:BASE
*/
static int i_open_file(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int n = 0;
   struct fd_def *fptr = &fd[fd_sys];
   struct rbf_def *rptr = &frbf[fd_sys];
   RELATION *tbptr;                       /* temporary R:BASE table pointer */

   while (fptr->flds[n].fldname[0])                     /* find field count */
      n++;
   rptr->fld_cnt = n;

/*
 *  if (fptr->openmode == OUTPUT_MODE)
 *  {
 *     errmsg("Creation of a new table file must be done with R:BASE");
 *     return(IOERROR);
 *  }
 */

   tbptr = rbopnrel(dbptr, fptr->filname);         /* open the table "file" */
   if (Rberrcod == 5) /* && (fptr->openmode != READ_MODE) */
   {
      errmsg("Creation of a new table file must be done with R:BASE");
      return(IOERROR);
   }
   if (Rberrcod)
      return(io_xlate(fd_sys, Rberrcod, "RB OPEN"));

   frbf[fd_sys].fdptr = tbptr;           /* the file is successfully opened */

   return(length_check(fd_sys));
}

/*
 *  Length check - checks actual record length against PRO-C's expectation
*/
static int length_check(fd_sys)
int fd_sys;
{
   int i, tlen, len = 0;
   struct fd_def *fptr = &fd[fd_sys];
   struct rbf_def *rptr = &frbf[fd_sys];
   RELINFO *infoptr;

   for (i = 0; i < rptr->fld_cnt; i++)        /* R:BASE length of PRO-C rec */
   {
      switch (fptr->flds[i].fldtype)
      {
         case LOGTYP :
         case MEMTYP :
         case CHRTYP :
            tlen = fptr->flds[i].fldlen;
            len += (tlen < 4 ? 4 : ((tlen % 2) ? tlen+1 : tlen));
            break;
         case DATTYP :
         case INTTYP :
         case LNGTYP :
            len += sizeof(long);
            break;
         case FLTTYP :
            len += sizeof(float);
            break;
         case DBLTYP :
            len += sizeof(double);
            break;
      }
   }

   infoptr = rbrelinf(dbptr, fptr->filname);
   tlen = infoptr->rowwid;                  /* row length as seen by R:BASE */
   free(infoptr);

   if (tlen != len)
   {
      rbclose(dbptr);
      errmsg(FileRecLenChg_s, fptr->filname);
      return(IOERROR);
   }
   return(IOGOOD);
}

/*
 *  File close function
 *  - if all tables are closed, it closes the overall database as well
*/
static int i_close_file(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int i, inactcnt = 0;

   rbclsrel(frbf[fd_sys].fdptr);
   if (Rberrcod)
      return(io_xlate(fd_sys, Rberrcod, "CLOSE"));

   fd[fd_sys].active = NO;

   for (i = 0; i < MAX_FILES; i++)
      if (fd[i].active == NO)
         inactcnt++;
   if (inactcnt == MAX_FILES)
      rbclose(dbptr);

   return(IOGOOD);
}


/*
 *  Select an index to perform processing on - already done in IOGEN.C
*/
static int i_selectinx(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
 *  Find a record by key value
*/
static int i_findkey(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int fx;
   struct fd_def *fptr = &fd[fd_sys];
   struct keyinfo *kptr = &fd[fd_sys].keys[fd[fd_sys].cur_key];

   fx = kptr->fldindex[0];
   setupkey(fd_sys, buffer, keybuff);

   strcpy(wherebuff, "where ");
   strcat(wherebuff, fptr->flds[fx].fldname);
   if (fd[fd_sys].exact)
      strcat(wherebuff, " EQ ");
   else
      strcat(wherebuff, " GE ");
   strcat(wherebuff, keybuff);

   rbwhere(frbf[fd_sys].fdptr, wherebuff);
   if (Rberrcod)
      return(io_xlate(fd_sys, Rberrcod, "RB FIND"));

   return(get_record(fd_sys, buffer));
}

/*
 *  Set up target key 
 *  - sets up a key string for record retrieval on current index
 *  - key search available only on first segment of a key at the moment.
*/
void setupkey(fd_sys, recbuf, keybuf)
int fd_sys;
char *recbuf, *keybuf;
{
   int i, kofs, klen, fx;
   long ltmp;
   float ftmp;
   double dtmp;
   struct fd_def *fptr = &fd[fd_sys];
   struct keyinfo *kptr = &fd[fd_sys].keys[fd[fd_sys].cur_key];

   zerorec(keybuf, MAX_FLDLEN);

   for (i = 0; i < kptr->segcount; i++)
   {
      kofs = kptr->segstart[i];
      klen = kptr->seglen[i];
      fx = kptr->fldindex[i];

      switch(fptr->flds[fx].fldtype)
      {
         case DATTYP :
            break;
         case CHRTYP :
            strcpy(keybuf, "\"");
            strcat(keybuf, &recbuf[kofs]);
            strcat(keybuf, "\"");
            break;
         case INTTYP :
         case LNGTYP :
            bytecpy(&ltmp, &recbuf[kofs], sizeof(long));
            fstrcpy(keybuf, fmt_lng(ltmp, "zzzzzzzz9"), 9);
            break;
         case FLTTYP :
            bytecpy(&ftmp, &recbuf[kofs], klen);
            dtmp = (double)ftmp;
            fstrcpy(keybuf, fmt_dbl(dtmp, "zzzzzzzzz.zzzzz"), 15);
            break;
         case DBLTYP :
            bytecpy(&dtmp, &recbuf[kofs], klen); 
            fstrcpy(keybuf, fmt_dbl(dtmp, "zzzzzzzzz.zzzzz"), 15);
            break;
      }
      if (kptr->segcount > 1)
         break;                 /* no support for segmented keys at present */
   }
}

/*
 *  Find first record in the file
*/
static int i_firstkey(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int fx;
   struct fd_def *fptr = &fd[fd_sys];
   struct keyinfo *kptr = &fd[fd_sys].keys[fd[fd_sys].cur_key];

   fx = kptr->fldindex[0];

   strcpy(wherebuff, "where ");
   strcat(wherebuff, fptr->flds[fx].fldname);
   strcat(wherebuff, " exists");

   rbwhere(frbf[fd_sys].fdptr, wherebuff);
   if (Rberrcod)
      return(io_xlate(fd_sys, Rberrcod, "RB FRST"));

   return(get_record(fd_sys, buffer));
}

/*
 *  Find last physical record in the file
*/
static int i_lastkey(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   strcpy(wherebuff, "where count EQ last");

   rbwhere(frbf[fd_sys].fdptr, wherebuff);
   if (Rberrcod)
      return(io_xlate(fd_sys, Rberrcod, "RB LAST"));

   return(get_record(fd_sys, buffer));
}

/*
 *  Find next record in the file
*/
static int i_nextrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   rbnxtrow(frbf[fd_sys].fdptr);
   if (Rberrcod)
      return((Rberrcod == -1) ? IOEOF : io_xlate(fd_sys, Rberrcod, "RB NEXT"));

   return(get_record(fd_sys, buffer));
}

/*
 *  Find previous record in the file
 *  - no analogous function to "next" in R:BASE interface library
*/
static int i_prevrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Retrieve a record - record locking is taken care of here
 *  - data is converted from R:BASE form to PRO-C buffer form
*/
static int get_record(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   int ofs, tlen, i, so = 0;
   int itmp;
   long ltmp;
   struct fd_def *fptr = &fd[fd_sys];

   rbgetrow(frbf[fd_sys].fdptr, storbuff);
   if (Rberrcod == -1)
      return(IOEOF);
   if (Rberrcod)
      return(io_xlate(fd_sys, Rberrcod, "RB GETREC"));

   zerorec(buffer, fptr->rec_len);
   
   for (i = 0; i < frbf[fd_sys].fld_cnt; i++)
   {
      ofs = fptr->flds[i].fldstart;
      tlen = fptr->flds[i].fldlen;

      switch(fptr->flds[i].fldtype)
      {
         case DATTYP :
            /* convert from long int back to string date */
            break;
         case LOGTYP :
         case MEMTYP :
         case CHRTYP :
            strncpy(&buffer[ofs], &storbuff[so], tlen);
            if (tlen < 4)
               so += 4;
            else if (tlen % 2 != 0)
               so += tlen+1;
            else
               so += tlen;                         /* offset for next field */
            break;
         case INTTYP :
            bytecpy(&ltmp, &storbuff[so], sizeof(long));
            itmp = (int)ltmp;
            bytecpy(&buffer[ofs], &itmp, sizeof(int));
            so += sizeof(long);
            break;
         case LNGTYP :
         case FLTTYP : 
         case DBLTYP :
            bytecpy(&buffer[ofs], &storbuff[so], tlen);
            so += tlen;
            break;
      }
   }
   return(IOGOOD);
} 


/*
 *  Re-read/reposition record pointer function - if required
*/
static int i_rereadrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(i_findkey(fd_sys, buffer));
}


/*
 *  Add a new record.
*/
static int i_addrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   buff2rb(fd_sys, buffer, storbuff);

   rbaddrow(frbf[fd_sys].fdptr, storbuff);
   return(Rberrcod ? io_xlate(fd_sys, Rberrcod, "RB ADD") : IOGOOD);
}

/*
 *  Update the current record.
*/
static int i_updrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   buff2rb(fd_sys, buffer, storbuff);

   if (i_lockrec(fd_sys, buffer) != IOGOOD)
      return(IOERROR);

   rbputrow(frbf[fd_sys].fdptr, storbuff);
   return(Rberrcod ? io_xlate(fd_sys, Rberrcod, "UPDREC") : IOGOOD);
}

/*
 *  Delete the current record.
*/
static int i_delrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   if (i_lockrec(fd_sys, buffer) != IOGOOD)
      return(IOERROR);

   rbdelrow(frbf[fd_sys].fdptr);
   return(Rberrcod ? io_xlate(fd_sys, Rberrcod, "DELREC") : IOGOOD);
}


/*
 *  Lock Record
*/  
static int i_lockrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   char *locklst[20];

   locklst[0] = frbf[fd_sys].fdptr;
   locklst[1] = (char *)0;

   rblock(locklst);
   return(Rberrcod ? io_xlate(fd_sys, Rberrcod, "LCKREC") : IOGOOD);
}

/*
 *  Unlock Record
*/
static int i_unlock_rec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   char *locklst[20];

   locklst[0] = frbf[fd_sys].fdptr;
   locklst[1] = (char *)0;

   rbunlock(locklst);
   return(Rberrcod ? io_xlate(fd_sys, Rberrcod, "ULCKREC") : IOGOOD);
}

/*
 *  Login
*/
static int i_login(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Logoff
*/
static int i_logoff(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  End (Commit) transaction 
*/
static int i_commit(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Rollback transaction
*/
static int i_rollback(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Start transaction
*/
static int i_transact(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
 *  Convert record to PRO-C buffer format from R:BASE form
*/
void buff2rb(fd_sys, P_buff, RB_buff)
int fd_sys;
char *P_buff, *RB_buff;
{
   int ofs, tlen, i, so = 0;
   int itmp;
   long ltmp;
   struct fd_def *fptr = &fd[fd_sys];

   zerorec(RB_buff, MAX_RECLEN);                /* blank out storage buffer */
   
   for (i = 0; i < frbf[fd_sys].fld_cnt; i++)
   {
      ofs = fptr->flds[i].fldstart;
      tlen = fptr->flds[i].fldlen;

      switch(fptr->flds[i].fldtype)
      {
         case MEMTYP :
         case LOGTYP :
         case CHRTYP :
            strncpy(&RB_buff[so], &P_buff[ofs], tlen);
            so += tlen;                            /* offset for next field */
            if (tlen < 4)
            {
               strncpy(&RB_buff[so], "    ", 4-tlen);
               so += 4-tlen;
            }
            else if (tlen % 2 != 0)
            {
               strcpy(&RB_buff[so], " ");
               so++;
            }
            break;
         case DATTYP :
            /* convert to long int date from Dec 31 4000 B.C. */
            /* ref. page 2-4 R:BASE program interface manual */
            /* return it in ltmp, a long integer */
            bytecpy(&RB_buff[so], &ltmp, sizeof(long));
            so += sizeof(long);
            break;
         case INTTYP :
            bytecpy(&itmp, &P_buff[ofs], tlen);
            ltmp = (long)itmp;
            bytecpy(&RB_buff[so], &ltmp, sizeof(long));
            so += sizeof(long);
            break;
         case LNGTYP :
         case FLTTYP :
         case DBLTYP :
            bytecpy(&RB_buff[so], &P_buff[ofs], tlen); 
            so += tlen;
            break;
      }
   }
}

/*
 *  This routine translates R:BASE Error codes into PRO-C error codes.
 *  If no PRO-C equivalent, displays the error number.
*/
static int io_xlate(fd_sys, ernum, rtnname)
int fd_sys;
int ernum;
char *rtnname;
{
   switch(ernum)
   {
      case -1    :
         return IONOKEY;
      case 1     :
      case 5     :
         return IONOFILE;
      case 3     : 
      case 511   :
         return IOBADOPEN;
   }

   if (fd_sys >= 0)
      errmsg(FileDbgError_sdss, "R:Base", ernum, fd[fd_sys].filname, rtnname);
   else
      errmsg(FileDbgError_sds, "R:Base", ernum, rtnname);
   return(IOERROR);
}


/*
 * Assign section
*/

void assign_IO_RB(dbnum)
int dbnum;
{
   Fntab[dbnum - 1][0]  = (int(*)())0; /* Empty */
   Fntab[dbnum - 1][1]  = i_init_file;
   Fntab[dbnum - 1][2]  = i_open_file;
   Fntab[dbnum - 1][3]  = i_close_file;
   Fntab[dbnum - 1][4]  = i_addrec;
   Fntab[dbnum - 1][5]  = i_delrec;
   Fntab[dbnum - 1][6]  = i_findkey;
   Fntab[dbnum - 1][7]  = i_firstkey;
   Fntab[dbnum - 1][8]  = i_lastkey;
   Fntab[dbnum - 1][9]  = i_lockrec;
   Fntab[dbnum - 1][10] = i_nextrec;
   Fntab[dbnum - 1][11] = i_prevrec;
   Fntab[dbnum - 1][12] = i_unlock_rec;
   Fntab[dbnum - 1][13] = i_updrec;
   Fntab[dbnum - 1][14] = i_commit;
   Fntab[dbnum - 1][15] = i_login;
   Fntab[dbnum - 1][16] = i_logoff;
   Fntab[dbnum - 1][17] = i_rollback;
   Fntab[dbnum - 1][18] = i_transact;
   Fntab[dbnum - 1][19] = i_selectinx;
   Fntab[dbnum - 1][20] = i_rereadrec;
   Fntab[dbnum - 1][21] = i_filename;
}
