/* ==( io/src/iuserdef.c )== */

/* ----------------------------------------------- */
/* Pro-C  Copyright (C) 1988 - 1990 Vestronix Inc. */
/* Modification to this source is not supported    */
/* by Vestronix Inc.                               */
/*            All Rights Reserved                  */
/* ----------------------------------------------- */
/* Written   NA   24-Feb-88                        */
/* Modified  xxx  DD-Mon-YY                        */
/* ----------------------------------------------- */
/* %W%  (%H% %T%) */

/*
 *  Modifications
 *
 *  DD-Mon-YY  xxx - dates, modification history...
*/

/*
 *  User Defined I/O module SKELETON via general IOGEN.C interface 
*/

/*
 *  This module handles the interface between the i/o calls from within
 *  PRO-C and a file of <user-defined file manager format>
 *  with fixed length records, containing fields of pre-defined length.
 *  Record, field, and key characteristics are defined in the .TH files
 *  generated by PRO-C and included or merged in the application source.
*/

/*
 *  All functions called through the function table take as arguments
 *  (int, char *) - usually index into array of file info structures, and
 *  the PRO-C data buffer passed from generated programs.  
 *  All should return either IOGOOD or a negative int PRO-C error code.
*/

# include <iodef.h>
# include <iomsg.h>
# include <proc.io>
# include <bench.h>
# include <iosup.h>

/* Function prototypes */
# ifdef ANSI
static int i_addrec(int, char *);
static int i_close_file(int, char *);
static int i_commit(int, char *);
static int i_delrec(int, char *);
static int i_filename(int, char *);
static int i_findkey(int, char *);
static int i_firstkey(int, char *);
static int i_init_file(int, char *);
static int i_lastkey(int, char *);
static int i_lockrec(int, char *);
static int i_login(int, char *);
static int i_logoff(int, char *);
static int i_nextrec(int, char *);
static int i_open_file(int, char *);
static int i_prevrec(int, char *);
static int i_rereadrec(int, char *);
static int i_rollback(int, char *);
static int i_selectinx(int, char *);
static int i_transact(int, char *);
static int i_unlock_rec(int, char *);
static int i_updrec(int, char *);
# else
static int i_addrec();
static int i_close_file();
static int i_commit();
static int i_delrec();
static int i_filename();
static int i_findkey();
static int i_firstkey();
static int i_init_file();
static int i_lastkey();
static int i_lockrec();
static int i_login();
static int i_logoff();
static int i_nextrec();
static int i_open_file();
static int i_prevrec();
static int i_rereadrec();
static int i_rollback();
static int i_selectinx();
static int i_transact();
static int i_unlock_rec();
static int i_updrec();
# endif


/*
 *  Declare any externs required by the file manager for error codes etc.
 *  Also declare any local variables/values required by the file manager 
 *  and applicable identically to all files opened.
*/

/*
 *  The usrdf_def structure should be used for specific values required by
 *  the file manager for each of the files to be opened, but not already
 *  included in the fd_def[] structure defined in iodef.h, which contains
 *  variables to hold file characteristics used by "most" file managers.
 *  The fd_sys parameter used throughout this module is the index into
 *  both the general fd_def[] array and the specific usrdf_def[] array
 *  declared here, with both structure arrays accessed "in parallel".
 *  Both usrdf_def and ufd[] should be renamed to avoid confusion if 
 *  several user defined I/O modules are being added to the library (see 
 *  existing PRO-C I/O modules for examples).
*/
struct usrdf_def
{
};

static struct usrdf_def ufd[MAX_FILES];



/*
 *
 * Interface Functions
 *
*/


/*
 *  Sets up File Name extension or strips the extension, as required by
 *  the file manager.
 *  File (or table) name is passed in buffer.
*/
static int i_filename(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  File and overall database initialization.
 *  This function will be called only once for each sequence of opening
 *  a file or set of files; it will be required again if the files are
 *  closed and then reopened.  PRO-C applications call the init_file() 
 *  function in iogen.c, which calls this, prior to open_file() calls.
 *  For file managers using an overall database structure that must be
 *  opened before individual tables are opened, this operation should be
 *  included here.  Other global initializations required before files
 *  are opened can also be put here.
 *  The overall database name, if required, is passed in the buffer.
*/
static int i_init_file(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  File (or table) open function.
 *  The basic set of required parameters is placed in the fd structure
 *  in the open_file() function in iogen.c before this specific open_file()
 *  function is called.  These include record length, field names, types,
 *  and lengths, key information, and file open modes.
 *  Any additional parameters for the file should be loaded into the
 *  parallel usrdf_def array element here.
 *  The i_open_file() function should perform a file open, or a create if
 *  the file does not exist.  Also, the index should be opened or created
 *  here, or through a local function called from this function.  Setting
 *  of update/read-only/output, and share/exclusive parameters for the
 *  file managers is also done here.
 *  The buffer parameter is a dummy argument here.
*/
static int i_open_file(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  If required, local file creation, key loading, and index creation 
 *  functions can be included here.  The argument lists can be whatever
 *  is required by the file manager; functions like this are not used 
 *  directly by PRO-C generated applications.
*/


/*
 *  File close function.
 *  For file managers that use an overall database file in addition to the 
 *  individually opened tables, this function should test for closure of
 *  all tables (a check of fd[n].active == YES can be used) and it should 
 *  issue a call to close the overall database, once all the individual 
 *  tables have been closed.  Because of the placement in the close_file()
 *  function in iogen.c of the assignment of fd[n].active = NO, it will be
 *  necessary to place this assignment in this specific function as well,
 *  after the file manager's close-table call, so that the check of all 
 *  tables having been closed will turn out correctly.
*/
static int i_close_file(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
 *  Select an index to perform processing on.
 *  This is already done in iogen.c, but the specific function allows for
 *  any additional processing required.
*/
static int i_selectinx(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
 *  Record retrieval functions.
 *  The current index is already set when these functions are called and
 *  the value in fd[fd_sys].cur_key represents the index to be used.
 *  Key values to search on are in the buffer at their specified field
 *  locations, key information including field offsets is in the 
 *  fd[fd_sys].keys[ck] structure member where ck = fd[fd_sys].cur_key.
 *  Key set-up, if required by the file manager, should be performed here.
 *  All of the record retrieval functions should exit with the record
 *  locked, if the retry parameter set in selectinx() in iogen.c indicates
 *  an active lock mode.
*/


/*
 *  Find a record by key value
 *  This replaces exactkey and partkey in PRO-C 1.3 Workbench library.
 *  The selectinx() function in iogen.c has already set fd[fd_sys].exact 
 *  to EXACT or PART depending on the value of the match parameter.
 *  This value should be used to select either an equal match or a GTE 
 *  match type function call from the file manager.
*/
static int i_findkey(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Find first record in the file.
 *  This is the first logical record in the file according to the current
 *  key.
*/
static int i_firstkey(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Find last physical record in the file.
 *  This is the last logical record in the file according to the current
 *  key.
 *  This function is used by auto-incrementing keys, and must return
 *  IOEOF with no error message if the file is empty.  If required, use
 *  an explicit check for an empty file.
*/
static int i_lastkey(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   if (** file is empty **)
      return(IOEOF);
      
   return(IOGOOD);
}

/*
 *  Find next record in the file
 *  The IOGEN function retains the original key value to use in checking
 *  for additional duplicate key records for an EXACT search.
 *  This function is required only to retrieve the record and return the 
 *  appropriate error status.
*/
static int i_nextrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Find previous record in the file
 *  The IOGEN function retains the original key value to use in checking
 *  for additional duplicate key records for an EXACT search.
 *  This function is required only to retrieve the record and return the 
 *  appropriate error status.
*/
static int i_prevrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
 *  Re-read/reposition record pointer function - if required.
 *  Used from IOGEN to re-establish record pointer after an unsuccessful
 *  retrieval function call.  fd[fd_sys].exact will be set to EXACT.
*/
static int i_rereadrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
 *  End of record retrieval functions.
 *  Local "helper" functions, for example, to set up a key, can be put
 *  here.
*/


/*
 *  Record write functions.
 *  All record write functions assume the record to be locked upon entry
 *  to the function via iogen.c, and the corresponding iogen.c functions
 *  unlock the record afterwards.
*/


/*
 *  Add a new record.
 *  Some file managers may require manual updating of indexes as well.
 *  Some may require explicit addition of a blank record first, or a 
 *  manual positioning to the end of file.
*/
static int i_addrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Update the current record.
 *  Some file managers may require explicit reseek or repositioning 
 *  before allowing update.  Some may require manual updating of indexes.
*/
static int i_updrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Delete the current record.
 *  Some file managers may require a record repositioning first.
 *  Some may require manual key deletion or manual index updating.
*/
static int i_delrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
 *  Lock Record
 *  Some file manager lock automatically on reads or writes, or allow a
 *  locking parameter to be included in record retrieval function calls.
 *  This function is used to carry out explicit file locking.  This can
 *  be done by a direct function call to a locking procedure, or a reread
 *  on the current record with the locking parameter included.
 *  This function will always be called from the iogen.c lock_rec() 
 *  function, whether fd[fd_sys].retry is set to NOLOCK or not; thus the
 *  NOLOCK value should be tested here.
*/  
static int i_lockrec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Unlock Record
 *  Some file managers unlock automatically.
 *  This function is used to carry out explicit file unlocking.  It is
 *  called directly by the file write functions in iogen.c.
 *  This function will always be called from the iogen.c ulock_rec() 
 *  function, whether fd[fd_sys].retry is set to NOLOCK or not; thus the
 *  NOLOCK value should be tested here.
*/
static int i_unlock_rec(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Login.
 *  This function is called from the init_file() function in iogen.c.
 *  If the file manager system requires user id and password, they are
 *  handled here.  The buffer contains the user id in the first 20 bytes,
 *  and the UNencrypted password in the second 20 bytes; these come from
 *  PRO-C's password module, if activated in the generated application.
 *  The administrator of the file manager used may have to set up matching
 *  accounts in advance to allow this "pass-through" of PRO-C passwords.
 *  If no passwords are activated through PRO-C, the buffer will contain
 *  NULL characters.
*/
static int i_login(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Logoff.
 *  This function could be used locally from the i_close_file() function 
 *  along with calls to close the entire database, where applicable.
*/
static int i_logoff(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  End (Commit) transaction.
*/
static int i_commit(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Rollback transaction.
*/
static int i_rollback(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}

/*
 *  Start transaction.
*/
static int i_transact(fd_sys, buffer)
int fd_sys;
char *buffer;
{
   return(IOGOOD);
}


/*
 *  Assign section.
 *  Existing file manager interface modules use two-letter codes for the
 *  file manager products supported by PRO-C, for example, PT for the 
 *  built-in PRO-TREE, CL for CodeBase with Clipper indexing, BT for
 *  Btrieve, etc.  (See the prototypes in iodef.h.)  The "UD" code below
 *  should be replaced with a two-letter mnemonic for the new file manager,
 *  and a prototype added to iodef.h.
*/
void assign_IO_UD(dbnum)
int dbnum;
{
   Fntab[dbnum - 1][0]  = (int(*)())0; /* Empty */
   Fntab[dbnum - 1][1]  = i_init_file;
   Fntab[dbnum - 1][2]  = i_open_file;
   Fntab[dbnum - 1][3]  = i_close_file;
   Fntab[dbnum - 1][4]  = i_addrec;
   Fntab[dbnum - 1][5]  = i_delrec;
   Fntab[dbnum - 1][6]  = i_findkey;
   Fntab[dbnum - 1][7]  = i_firstkey;
   Fntab[dbnum - 1][8]  = i_lastkey;
   Fntab[dbnum - 1][9]  = i_lockrec;
   Fntab[dbnum - 1][10] = i_nextrec;
   Fntab[dbnum - 1][11] = i_prevrec;
   Fntab[dbnum - 1][12] = i_unlock_rec;
   Fntab[dbnum - 1][13] = i_updrec;
   Fntab[dbnum - 1][14] = i_commit;
   Fntab[dbnum - 1][15] = i_login;
   Fntab[dbnum - 1][16] = i_logoff;
   Fntab[dbnum - 1][17] = i_rollback;
   Fntab[dbnum - 1][18] = i_transact;
   Fntab[dbnum - 1][19] = i_selectinx;
   Fntab[dbnum - 1][20] = i_rereadrec;
   Fntab[dbnum - 1][21] = i_filename;
}

