/* ==( memo/memo.c )== */

/* ----------------------------------------------- */
/* Pro-C  Copyright (C) 1989 - 1990 Vestronix Inc. */
/* Modification to this source is not supported    */
/* by Vestronix Inc.                               */
/*            All Rights Reserved                  */
/* ----------------------------------------------- */
/* Written   JPK   1-May-89                        */
/* Modified  Geo  12-Dec-89  See comments below    */
/* ----------------------------------------------- */
/* %W%  (%H% %T%) */

/*
 *  Modifications
 *
 *  11-Jan-90  JH  - word wrap
 *	20-Dec-89  JH  - rewrite, bug fixes
 *  12-Dec-89  Geo - V2 version
 *
*/

# include <stdio.h>
# include <bench.h>
# include "field.h"


# ifdef ANSI
static void disp_memo(int, int, int, int, int);
static void adjust_rows(int, int);
static void join_lines(int, int);
static void join_all_lines(int, int, int);
static int check_wrap(int, int, int);
char *forward(FIELD *, char *);
void home_field(FIELD *);
static void pad_out_buf(int, int);
static int endstroke(int);
# else
static void disp_memo();
static void adjust_rows();
static void join_lines();
static void join_all_lines();
static int check_wrap();
char *forward();
void home_field();
static void pad_out_buf();
static int endstroke();
# endif

#define MIN(a, b)	(a > b ? b : a)


extern int cursor_col;		/* inputt.c */
extern int instos;			/* inchar.c */
extern int field_overflow;	/* inputt.c */
extern int homed_field;		/* TRUE if field has been homed, so text_input
							* does not reset it */
extern int doing_memos;

static FIELD *fld_list;      /* pointer to memo line input list */



/*
 * Edit a memo. - that says it all, now doesn't it !!!!
*/
void memo_input(win_height, win_width, attr, nrows, help_num, memo_array)
	int win_height;
	int win_width;
	int attr;
	int nrows;
	int help_num;
	char *memo_array;
{
	int i;
	int exitcode;
	int done = FALSE;
	int cur_fld = 0;
	int top_fld = 0;
	FIELD *cur_field;
	char *memo_mask;      /* pointer to memo line mask */

	doing_memos = TRUE;

	/* check for non-postive numbers */
	if (win_height <= 0 || win_width <= 0 || nrows <= 0)
		return;
	
	if (nrows < win_height)
		win_height = nrows + 2;

	/* actual width of text is two less than window width */
	win_width -= 2;
	win_height -= 2;

	/* allocate the field input mask */
	if ((memo_mask = (char *)alloc(win_width+1)) == NULL)
		return;
	byteset(memo_mask, MASK_x, win_width);
	memo_mask[win_width] = '\0';

	/* allocate the field list */
	if ((fld_list = (FIELD *)alloc(sizeof(FIELD) * (nrows + 1))) == NULL)
		return;

	/* fill in the field list parameters, field by field */
	for (i = 0; i < nrows; i++)
	{
		(fld_list+i)->frow        = i + 2;
		(fld_list+i)->fcol        = 2;
		(fld_list+i)->fx          = 1;
		(fld_list+i)->fbuff       = memo_array + (i * win_width);
		(fld_list+i)->fmask       = memo_mask;
		(fld_list+i)->fhelp       = help_num;
		(fld_list+i)->fieldlen    = win_width;
		(fld_list+i)->ftype       = F_TEXT;
		(fld_list+i)->fattrib_on  = attr;
		(fld_list+i)->fattrib_off = attr;
		(fld_list+i)->fcontrol    = OVERFLOW_RETURN | NO_BLANK;
		(fld_list+i)->f_pre       = NULL;
		(fld_list+i)->f_input     = text_input;
		(fld_list+i)->f_val       = NULL;
		(fld_list+i)->f_post      = NULL;
	}

	cur_field = fld_list;

	/* do the original display */
	disp_memo(top_fld, top_fld, attr, win_height, win_width);

	while (!done)
	{
		exitcode = input_wx(cur_field, K_ESC, K_TAB, K_F2, K_F3, K_DEL, K_RIGHT, K_LEFT, K_UP, K_DOWN, K_PGUP, K_PGDN, K_HOME, K_END, 0);

		switch (exitcode)
		{
		case K_CR :
			if (instos == TRUE)
			{
				/* is last line null ? */
				if ((cur_fld != nrows-1) && *(fld_list+nrows-1)->fbuff == '\0')
				{
					int len;

				  	move_text((cur_field+2)->fbuff, (cur_field+1)->fbuff, win_width * (nrows - cur_fld - 2));

					/* split line at cursor position */
					len = strlen(cur_field->fbuff) - cursor_col + 1;

					byteset((cur_field+1)->fbuff, '\0', win_width);
					move_text((cur_field+1)->fbuff, cur_field->fbuff+cursor_col-2, len);
					byteset(cur_field->fbuff+cursor_col-2, '\0', len+1);
					disp_memo(top_fld, top_fld, attr, win_height, win_width);
				}	
			}
		case K_RIGHT :
		case K_DOWN :
			/* move to next field */
			if (cur_fld == nrows - 1)
				do_bell();
			else
			{
				cur_fld++;
				cur_field++;

				if ((cur_fld - top_fld) >= win_height)
				{
					top_fld++;
					adjust_rows(top_fld, win_height);
					disp_memo(top_fld, top_fld, attr, win_height, win_width);
				}
				home_field(cur_field);
				homed_field = TRUE;

				if (exitcode == K_DOWN)
					cur_field->fx = (cur_field-1)->fx > strlen(cur_field->fbuff) ? strlen(cur_field->fbuff) : (cur_field-1)->fx;
				else /* K_CR || K_RIGHT */
					cur_field->fx = 0;
			}
			break;

		case K_LEFT :
		case K_UP :
			if (cur_fld == 0)
				do_bell();
			else
			{
				--cur_fld;
				--cur_field;

				if (cur_fld < top_fld)
				{
					top_fld = cur_fld;
					adjust_rows(top_fld, win_height);
					disp_memo(top_fld, top_fld, attr, win_height, win_width);
				}
				home_field(cur_field);
				homed_field = TRUE;

				if (exitcode == K_UP)
					cur_field->fx = MIN((cur_field+1)->fx, strlen(cur_field->fbuff));
				else /* K_LEFT */
					cur_field->fx = MIN(strlen(cur_field->fbuff), cur_field->fieldlen);
			}
			break;

		case K_HOME :
			/* go to first line of memo */
			cur_fld = top_fld = 0;
			cur_field = fld_list;
			adjust_rows(top_fld, win_height);
			disp_memo(top_fld, top_fld, attr, win_height, win_width);
			break;

		case K_END :
			cur_fld = nrows - 1;
			cur_field = fld_list + nrows - 1;

			while (cur_fld > 0 && cur_field->fbuff[0] == '\0')
			{
				cur_fld--;
				cur_field--;
			}

			if ((top_fld = cur_fld - win_height + 1) < 0)
				top_fld = 0;
			adjust_rows(top_fld, win_height);
			disp_memo(top_fld, top_fld, attr, win_height, win_width);
			break;

		case K_PGUP :
			if (cur_fld == 0)
				do_bell();
			else
			{
				int save_fx = cur_field->fx;

				if ((cur_fld -= win_height - 1) < 0)
					cur_fld = 0;
				cur_field = fld_list + cur_fld;
				if (cur_fld < top_fld)
				{
					top_fld = cur_fld;
					adjust_rows(top_fld, win_height);
					disp_memo(top_fld, top_fld, attr, win_height, win_width);
				}
				home_field(cur_field);
				homed_field = TRUE;

				cur_field->fx = save_fx > strlen(cur_field->fbuff) ? strlen(cur_field->fbuff) : save_fx;
			}
			break;

		case K_PGDN :
			if (cur_fld == nrows - 1)
				do_bell();
			else
			{
				int save_fx = cur_field->fx;
				
				if ((cur_fld += win_height - 1) > nrows - 1)
					cur_fld = nrows - 1;
				cur_field = fld_list + cur_fld;

				/* adjust the top field for memo display */
				if ((cur_fld - top_fld) >= win_height)
				{
					top_fld = cur_fld - (win_height - 1);
					adjust_rows(top_fld, win_height);
					disp_memo(top_fld, top_fld, attr, win_height, win_width);
				}
				home_field(cur_field);
				homed_field = TRUE;

				cur_field->fx = save_fx > strlen(cur_field->fbuff) ? strlen(cur_field->fbuff) : save_fx;
			}
			break;
		
		case K_DEL :
			if ( *(cur_field->fbuff) == '\0')
			{
				move_text(cur_field->fbuff, (cur_field+1)->fbuff, win_width * (nrows - cur_fld));
				byteset((fld_list+nrows-1)->fbuff, '\0', win_width);
			}
			disp_memo(cur_fld, top_fld, attr, win_height, win_width);
			break;

		case K_F2 :
			/* join this line with the one below it */
			join_lines(cur_fld, win_width);
			if ( *((cur_field+1)->fbuff) == '\0')
			{
				move_text((cur_field+1)->fbuff, (cur_field+2)->fbuff, win_width * (nrows - cur_fld - 1));
				byteset((fld_list+nrows-1)->fbuff, '\0', win_width);
			}
			disp_memo(cur_fld, top_fld, attr, win_height, win_width);
			{
				int tmp_fx = cur_field->fx;

				home_field(cur_field);
				cur_field->fx = tmp_fx;
				homed_field = TRUE;
			}
			break;

		case K_F3 :
			/* join all lines */
			join_all_lines(cur_fld, win_width, nrows);
			disp_memo(cur_fld, top_fld, attr, win_height, win_width);
			break;

		default :
			if (field_overflow)
			{
				int wrap; 

				if (exitcode == HARDSPACE)
				{
					cur_fld++;
					cur_field++;

					if ((cur_fld - top_fld) >= win_height)
					{
						top_fld++;
						adjust_rows(top_fld, win_height);
					}
					disp_memo(top_fld, top_fld, attr, win_height, win_width);
					break;
				}

				/* do the wrap thing */
				if ((wrap = check_wrap(cur_fld, win_width, nrows)) != 0)
				{
					cur_fld++;
					cur_field++;

					if ((cur_fld - top_fld) >= win_height)
					{
						top_fld++;
						adjust_rows(top_fld, win_height);
					}

					home_field(cur_field);
					homed_field = TRUE;

					*(cur_field->fbuff+wrap-1) = (char)exitcode;
					cur_field->fx = wrap > strlen(cur_field->fbuff) ? strlen(cur_field->fbuff) : wrap;
				}
				disp_memo(top_fld, top_fld, attr, win_height, win_width);
				break;
			}
			else
				done = endstroke(exitcode);
			break;
		}
	}

	/*
	* should pad out the buffer with spaces 
	*/
	pad_out_buf(nrows, win_width);

	free(fld_list);
	free(memo_mask);
}


/*
 * Test to see if a character is an input ending character
*/
static int endstroke(c)
	int c;
{
	if (c == K_ESC)
		return TRUE;
	return FALSE;
}


/*
 * Display lines of a memo starting from the start and continuing until 
 * the bottom of the memo window is reached.
*/
static void disp_memo(curfield, topfield, attr, height, ncols)
	int curfield;
	int topfield;
	int attr;
	int height;
	int ncols;
{
	int i;
	FIELD *fld_ptr;

	fld_ptr = fld_list + curfield;
	for (i = 0; i < height-curfield+topfield; i++, fld_ptr++)
		ndisp_w(fld_ptr->frow, fld_ptr->fcol, attr, ncols, fmt_text(fld_ptr));
}


/*
 * Set the display row for all memo lines that are currently being displayed
 * in the memo window starting with the line that is the current top line 
*/
static void adjust_rows(top_fld, height)
	int top_fld;
	int height;
{
	int i = 0;

	while (i < height)
	{
		(fld_list+top_fld+i)->frow = 2 + i;
		i++;
	}
}



/*
 * Join the current line of the memo with the next line and wrap the overflow
 * back to the next line; if all_lines is TRUE then carry out the joining for
 * every line of the mehich will have the effect of reformatting the memo.
*/
static void join_lines(cur_fld, ncols)
	int cur_fld;
	int ncols;
{
	int line_no;
	char *line_buf;
	char *bptr;
	FIELD *cur_field;
	int len;

	/* allocate a buffer to use in joining two lines together */
	if ((line_buf = (char *)alloc(2* (ncols + 1))) == NULL)
		return;
	
	line_no = cur_fld;
	cur_field = fld_list+line_no;

	strncpy(line_buf, cur_field->fbuff, ncols);
	if ((len = strlen(line_buf)) > 0)
	{
		line_buf[len] = HARDSPACE;
		strncat(line_buf, (cur_field+1)->fbuff, ncols);
	}
	else
		strncpy(line_buf, (cur_field+1)->fbuff, ncols);

	/* now pull off the maximum that will fit in a line */
	if (strlen(line_buf) > ncols)
	{
		bptr = line_buf + strlen(line_buf);
		do 
		{
			while (bptr != line_buf && *bptr != HARDSPACE)
				bptr--;
			if ((bptr - line_buf) <= ncols)
			{
				*bptr = '\0';
				strncpy(cur_field->fbuff, line_buf, ncols);
				strcpy((cur_field+1)->fbuff, bptr+1);
				break;
			}
			else
				bptr--;
		} while (bptr >= line_buf);
	}
	else
	{
		strncpy(cur_field->fbuff, line_buf, ncols);
		(cur_field+1)->fbuff[0] = '\0';
	}

	free(line_buf);
}


static void join_all_lines(cur_fld, ncols, nrows)
	int cur_fld;
	int ncols;
	int nrows;
{
	char *line_buf;
	char *bptr;
	FIELD *cur_field;
	int i;
	int line_no;
	char *start_ptr;

	if ((line_buf = (char *)alloc((nrows-cur_fld+1)*ncols)) == NULL)
		return;
	
	line_no = cur_fld;
	cur_field = fld_list+cur_fld;
	i = cur_fld;
	
	strncpy(line_buf, cur_field->fbuff, ncols);
	while( ++i < nrows-cur_fld)
	{
		strcat(line_buf, " ");
		strncat(line_buf, (cur_field+i)->fbuff, ncols);
	}

	byteset(cur_field->fbuff, '\0', (nrows-cur_fld)*ncols);

	bptr = line_buf;
	start_ptr = bptr;
	i=0;
	while (*bptr)
	{
		i++;
		bptr++;
		if (i == ncols)
		{
		  while (!isspace(*bptr))
		  {
			  bptr--;
			  i--;
		  }
		  while (isspace(*bptr))
		  {
			  bptr--;
			  i--;
		  }
			/* be pointing to the space after the string */
			i++;
			bptr++;

			if (i>0)
			{
				if (i + strlen((fld_list+line_no)->fbuff) > ncols)
					line_no++;
				strncpy((fld_list+line_no)->fbuff+strlen((fld_list+line_no)->fbuff), start_ptr, i);
			}
			while (isspace(*bptr))
				bptr++;

			start_ptr = bptr;
			i=0;
		}		
	}
	/* put the last chars back into memo */
	if (*start_ptr)
	{
	 	if (i + strlen((fld_list+line_no)->fbuff) + 1 > ncols)
			 line_no++;
		strncpy((fld_list+line_no)->fbuff+strlen((fld_list+line_no)->fbuff), start_ptr, ncols);
	}
}



/*
 * Check the current line of the memo to see if wordwrap is needed; if it is
 * then the routine calls itself recursive to make sure enough room is left
 * below it to handle the wrap.
*/
static int check_wrap(cur_fld, ncols, nrows)
	int cur_fld;
	int ncols;
	int nrows;
{
	char *word_ptr;
	char *bptr;
	int wraplen = 0;
	int tmp_wrap_len;
	FIELD *cur_field = fld_list + cur_fld;

	/* is there a next line */
	if (cur_fld == nrows - 1)
		return 0;

	/* locate the start and size of the last word */
	bptr = cur_field->fbuff + cur_field->fieldlen;
	while (*(--bptr) != HARDSPACE && bptr != cur_field->fbuff)
		wraplen++;

	if (wraplen >= ncols - 1)
		return 1;

	if (*bptr == HARDSPACE)
		bptr++;
	word_ptr = bptr;

	if (strlen((cur_field+1)->fbuff) > (ncols - wraplen))
		return 1;

	/* put the word on the next line and fill old spot with spaces */
	move_text((cur_field+1)->fbuff+wraplen+1, (cur_field+1)->fbuff, ncols-wraplen-2);
	move_text((cur_field+1)->fbuff, word_ptr, wraplen);

	tmp_wrap_len = wraplen;
	while (tmp_wrap_len--)
		*word_ptr++ = '\0';

	return wraplen + 1;
}


static void pad_out_buf(nrows, ncols)
	int nrows;
	int ncols;
{
	int i;
	int j;
	char *buf_ptr;

	for (i=0; i < nrows; i++ )	
	{
		buf_ptr = (fld_list + i)->fbuff;

		for (j=0; j < ncols; j++)
		{
			if (!*buf_ptr)
				*buf_ptr=' ';
			buf_ptr++;
		}
	}
	*buf_ptr='\0';

	while (*(--buf_ptr) == ' ');
	*(++buf_ptr)='\0';
}



