/* ==( bench/msg.c )== */
/* ----------------------------------------------- */
/* Pro-C  Copyright (C) 1988 - 1990 Vestronix Inc. */
/* Modification to this source is not supported    */
/* by Vestronix Inc.                               */
/*            All Rights Reserved                  */
/* ----------------------------------------------- */
/* Written   Geo  26-Aug-88                        */
/* Modified  Geo   8-Mar-90  See comments below    */
/* ----------------------------------------------- */
/* %W%  (%H% %T%) */

/*
 *  Modifications
 *
 *   8-Mar-90  Geo  Variable sized windies and tighted up code
 *  20-Feb-90  Nig  Warning/Query call option with -ve help no.
 *   4-Jan-90  Geo  Added higlighter to errmsgs using ^ method
 *  13-Dec-89  Geo/BF combined errmsg, warn & working plus NEW
 *   statmsg which really does a working (but deletes itself)
 *  12-Dec-89  Geo - V2 version with variable lines
 *  25-Oct-89  Geo - 1.32 Merge
 *
 *
*/
/* Still to do
 * word wrapping & dynamic allocation - problem is how much approx 1/2K ?
 * also merging this sub functions so far 2 are stderr 1 is synch
*/

# include <stdio.h>
# include <bench.h>

extern int _wsystart;      /* win.c */

/* Function prototypes */
# ifdef ANSI
static void get_size(char *, int *, int *, int *, int *);
static void show_buf(int *, int *, char *);
# else
static void get_size();
static void show_buf();
# endif

# define MAX_BUFF 512 /* Increase if stack can take it */

# ifdef UNIX
int errmsg(va_alist)
va_dcl
# else
int errmsg(char *va_alist, ...)
# endif
{
	va_list ap;
	char *fmt;

   /*
    * Vararg section for status box
   */
# ifdef UNIX 
	va_start(ap);
	fmt = va_arg(ap, char *);
# else
	va_start(ap, va_alist);
	fmt = va_alist;
# endif

	/* Use windowing if enabled otherwise stderr */
	if (_wsystart)
	{
	char dispbuff[MAX_BUFF];   /* use dynamic later (Geo) */
	int ch;
	int height = 5, width = 2;

		vsprintf(dispbuff, fmt, ap);

		show_buf(&height, &width, dispbuff);

		center_w(height - 1, 1, NORMAL, width, ok_prompt);
		moveto_w(height - 1, width/2);

		while (ch = inchar())
			if (ch == K_CR || ch == K_ESC || ch == M_RELEASE ||
					ch == 'o' || ch == 'O')
				break;

		delete_w();
	}
	else
	{
		/* Direct support for batch calls */
		vfprintf(stderr, fmt, ap);
		fputc('\n', stderr);
		fflush(stderr);
	}

	va_end(ap);

	return(TRUE); /* Historic - but leave alone ! */
}


# ifdef UNIX
void statmsg(va_alist)
va_dcl
# else
void statmsg(char *va_alist, ...)
# endif
{
	va_list ap;
	char *fmt;

   /*
    * Vararg section for status box
   */
# ifdef UNIX 
	va_start(ap);
	fmt = va_arg(ap, char *);
# else
	va_start(ap, va_alist);
	fmt = va_alist;
# endif

	/* Use windowing if enabled otherwise stderr */
	if (_wsystart)
	{
	char dispbuff[MAX_BUFF];   /* use dynamic later (Geo) */
	int height = 3, width = 1;

		vsprintf(dispbuff, fmt, ap);

		show_buf(&height, &width, dispbuff);

		flushscr();
		delete_w();    /* Will disappear on next flush (NEAT, HUH?) */
	}
	else
	{
		/* Direct support for batch calls */
		vfprintf(stderr, fmt, ap);
		fputc('\n', stderr);
		fflush(stderr);
	}

	va_end(ap);
}

/*
 * Asks the user a Yes/No Question
*/
# ifdef UNIX
int warning(helpnum, va_alist)
int helpnum;
va_dcl
# else
int warning(int helpnum, char *va_alist, ...)
# endif
{
	char dispbuff[MAX_BUFF];   /* use dynamic later (Geo) */
	va_list ap;   /* Variable argument list for functions below */
	char *fmt;
	int ret = 0;
   static struct optab choices[] =
   {
      { 4, 10, yes_prompt },
      { 4, 15, no_prompt },
      { REVVID, NORMAL, NULL }
   };
	int height = 5, width, tmp;
	width = tmp = strlen(yes_prompt) + strlen(no_prompt) + 1; 
	tmp++;
	
	width = width < 6 ? 6 : width;

   /*
    * Vararg section for bottom status line
   */
# ifdef UNIX
	va_start(ap);
	fmt = va_arg(ap, char *);
# else
	va_start(ap, va_alist);
	fmt = va_alist;
# endif

	vsprintf(dispbuff, fmt, ap);
	va_end(ap);

	show_buf(&height, &width, dispbuff);

	choices[0].col = (width - tmp) / 2 + 1;
	choices[0].col = choices[0].col < 2 ? 2 : choices[0].col;
	choices[1].col = choices[0].col + strlen(yes_prompt) + 2;
	choices[0].row = choices[1].row = height - 1;

  	ret = do_options(choices, ret, -1);

	delete_w();

	/* Adjust for Yes/No */
	/* Only test for yes all else is No */
	return((ret == 0) ? 1 : 0);
}

/*
 * Asks the user a Yes/No/Cancel Question
*/
# ifdef UNIX
int query(helpnum, va_alist)
int helpnum;
va_dcl
# else
int query(int helpnum, char *va_alist, ...)
# endif
{
	char dispbuff[MAX_BUFF];   /* use dynamic later (Geo) */
	va_list ap;   /* Variable argument list for functions below */
	char *fmt;
	int ret = 0;
   static struct optab choices[] =
   {
      { 4,  5, yes_prompt },
      { 4, 10, no_prompt },
      { 4, 15, cancel_prompt },
      { REVVID, NORMAL, NULL }
   };
	int height = 5, tmp, width;
	/* width = text length only */
	width = tmp = strlen(yes_prompt) + strlen(no_prompt) + strlen(cancel_prompt) + 2;
	tmp += 2;
	width = width < 13 ? 13 : width;

   /*
    * Vararg section for bottom status line
   */
# ifdef UNIX
	va_start(ap);
	fmt = va_arg(ap, char *);
# else
	va_start(ap, va_alist);
	fmt = va_alist;
# endif

	vsprintf(dispbuff, fmt, ap);
	va_end(ap);

	show_buf(&height, &width, dispbuff);

	choices[0].col = (width - tmp) / 2 + 1;
	choices[0].col = choices[0].col < 2 ? 2 : choices[0].col;
	choices[1].col = choices[0].col + strlen(yes_prompt) + 2;
	choices[2].col = choices[1].col + strlen(no_prompt) + 2;

	choices[0].row = choices[1].row = choices[2].row = height - 1;

	ret = do_options(choices, ret, -1);

	if (ret != -1)
		ret = ( ret == 0 ) ? 1 : (ret == 2) ? -1 : 0; /* Adjust for Yes/No/Cancel */

	delete_w();
	return(ret);
}

/*
 * Formatting routines
*/
static void get_size(buf, row, col, height, width)
char *buf;
int *row, *col, *height, *width;
{
int lines = 0;
int len = 0;
int maxlen = *width;

	/* count # of lines */
	while (*buf)
	{
		if (*buf++ == '\n')
		{
			lines++;
			if (len > maxlen)
				maxlen = len;
			len = 0;
		}
		else
			if (*buf != '^') /* Added for toggle attr not both ^ on same line */
				len++;
	}
	if (len > maxlen)
		maxlen = len;

	/* Check widths and aggregate heights (widths) to avoid overflow */
	/* Only bother checking the width */
	if (maxlen > w_ncols - 4)
	{
		*height += (maxlen + w_ncols - 5) / (w_ncols - 4) - 1;
		maxlen = w_ncols - 4;
	}
	*width = maxlen + 4;
	*height += lines;

	*row -= *height/2 - 1;
	*col -= *width/2 - 1;
}

static void show_buf(height, width, buf)
int *height, *width;
char *buf;
{
int toggle = 0; /* Toggle REVVID and BOLD */
int row = w_nrows/2, col = w_ncols/2;

	get_size(buf, &row, &col, height, width);

	/* Init the window */
	create_w(row, col, *height, *width);
	set_w(REVVID, ' ');
	border_w(0, REVVID);

	/* Don't modify width & height affter setting once */

	/* Display stage */
	row = 2; col = 3;
	while (*buf)
	{
		if (*buf == '\n' || col > *width - 2)
		{
			row++;
			col = 3;
		}

		if (*buf != '\n')
		{
			if (*buf == '^')
				toggle = !toggle;
			else
				poke_w(row, col++, (toggle ? BOLD : REVVID), *buf);
		}
		buf++;
	}
}

